import warnings
from collections import defaultdict
from collections.abc import Collection, Hashable, Sequence
from typing import Any, cast

from cognite.client.data_classes import data_modeling as dm
from cognite.client.data_classes.data_modeling.containers import BTreeIndex
from cognite.client.data_classes.data_modeling.data_types import EnumValue as DMSEnumValue
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType
from cognite.client.data_classes.data_modeling.views import (
    SingleEdgeConnectionApply,
    SingleReverseDirectRelationApply,
    ViewPropertyApply,
)

from cognite.neat._client.data_classes.data_modeling import (
    ContainerApplyDict,
    NodeApplyDict,
    SpaceApplyDict,
    ViewApplyDict,
)
from cognite.neat._client.data_classes.schema import DMSSchema
from cognite.neat._constants import COGNITE_SPACES
from cognite.neat._issues.errors import NeatTypeError, NeatValueError, ResourceNotFoundError
from cognite.neat._issues.warnings import NotSupportedWarning, PropertyNotFoundWarning
from cognite.neat._issues.warnings.user_modeling import (
    EmptyContainerWarning,
    HasDataFilterOnNoPropertiesViewWarning,
)
from cognite.neat._rules.models.data_types import DataType, Double, Enum, Float
from cognite.neat._rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    DMSFilter,
    DMSNodeEntity,
    DMSUnknownEntity,
    EdgeEntity,
    HasDataFilter,
    NodeTypeFilter,
    ReverseConnectionEntity,
    UnitEntity,
    ViewEntity,
)

from ._rules import DMSEnum, DMSMetadata, DMSProperty, DMSRules, DMSView


class _DMSExporter:
    """The DMS Exporter is responsible for exporting the DMSRules to a DMSSchema.

    This kept in this location such that it can be used by the DMSRules to validate the schema.
    (This module cannot have a dependency on the exporter module, as it would create a circular dependency.)

    Args
        include_pipeline (bool): If True, the pipeline will be included with the schema. Pipeline means the
            raw tables and transformations necessary to populate the data model.
        instance_space (str): The space to use for the instance. Defaults to None,`Rules.metadata.space` will be used
        remove_cdf_spaces(bool): The
    """

    def __init__(self, rules: DMSRules, instance_space: str | None = None, remove_cdf_spaces: bool = False):
        self.instance_space = instance_space
        self.rules = rules
        self.remove_cdf_spaces = remove_cdf_spaces

    def to_schema(self) -> DMSSchema:
        rules = self.rules
        container_properties_by_id, view_properties_by_id = self._gather_properties(list(self.rules.properties))

        containers = self._create_containers(container_properties_by_id, rules.enum)  # type: ignore[arg-type]

        view_properties_with_ancestors_by_id = self._gather_properties_with_ancestors(
            view_properties_by_id, rules.views
        )

        views = self._create_views(view_properties_by_id, view_properties_with_ancestors_by_id)
        view_node_type_filters: set[dm.NodeId] = set()
        for dms_view in rules.views:
            if isinstance(dms_view.filter_, NodeTypeFilter):
                view_node_type_filters.update(node.as_id() for node in dms_view.filter_.inner or [])
        if rules.nodes:
            node_types = NodeApplyDict(
                [node.as_node() for node in rules.nodes]
                + [dm.NodeApply(node.space, node.external_id) for node in view_node_type_filters]
            )
        else:
            node_types = NodeApplyDict(
                [
                    dm.NodeApply(node.space, node.external_id)
                    for node in view_node_type_filters
                    if not (self.remove_cdf_spaces and node.space in COGNITE_SPACES)
                ]
            )

        data_model = rules.metadata.as_data_model()
        # Sorting to ensure deterministic order
        data_model.views = sorted(
            [dms_view.view.as_id() for dms_view in rules.views if dms_view.in_model],
            key=lambda x: x.as_tuple(),  # type: ignore[union-attr]
        )
        spaces = self._create_spaces(rules.metadata, containers, views, data_model)

        return DMSSchema(
            spaces=spaces,
            data_model=data_model,
            views=views,
            containers=containers,
            node_types=node_types,
        )

    def _create_spaces(
        self,
        metadata: DMSMetadata,
        containers: ContainerApplyDict,
        views: ViewApplyDict,
        data_model: dm.DataModelApply,
    ) -> SpaceApplyDict:
        used_spaces = (
            {container.space for container in containers.values()}
            | {view.space for view in views.values()}
            | {data_model.space}
            | {metadata.space}
        )

        spaces = SpaceApplyDict([dm.SpaceApply(space=space) for space in used_spaces])
        if self.instance_space and self.instance_space not in spaces:
            spaces[self.instance_space] = dm.SpaceApply(space=self.instance_space, name=self.instance_space)
        return spaces

    def _create_views(
        self,
        view_properties_by_id: dict[dm.ViewId, list[DMSProperty]],
        view_properties_with_ancestors_by_id: dict[dm.ViewId, list[DMSProperty]],
    ) -> ViewApplyDict:
        input_views = list(self.rules.views)

        views = ViewApplyDict(
            [
                dms_view.as_view()
                for dms_view in input_views
                if not (self.remove_cdf_spaces and dms_view.view.space in COGNITE_SPACES)
            ]
        )
        view_by_id = {dms_view.view: dms_view for dms_view in input_views}

        edge_types_by_view_property_id = self._edge_types_by_view_property_id(
            view_properties_with_ancestors_by_id, view_by_id
        )

        for view_id, view in views.items():
            view.properties = {}
            if not (view_properties := view_properties_by_id.get(view_id)):
                continue
            for prop in view_properties:
                view_property = self._create_view_property(
                    prop, view_properties_with_ancestors_by_id, edge_types_by_view_property_id
                )
                if view_property is not None:
                    view.properties[prop.view_property] = view_property

        return views

    @classmethod
    def _create_edge_type_from_prop(cls, prop: DMSProperty) -> dm.DirectRelationReference:
        if isinstance(prop.connection, EdgeEntity) and prop.connection.edge_type is not None:
            return prop.connection.edge_type.as_reference()
        elif isinstance(prop.value_type, ViewEntity):
            return cls._default_edge_type_from_view_id(prop.view.as_id(), prop.view_property)
        else:
            raise NeatTypeError(f"Invalid valueType {prop.value_type!r}")

    @staticmethod
    def _default_edge_type_from_view_id(view_id: dm.ViewId, property_: str) -> dm.DirectRelationReference:
        return dm.DirectRelationReference(
            space=view_id.space,
            # This is the same convention as used when converting GraphQL to DMS
            external_id=f"{view_id.external_id}.{property_}",
        )

    @classmethod
    def _edge_types_by_view_property_id(
        cls,
        view_properties_with_ancestors_by_id: dict[dm.ViewId, list[DMSProperty]],
        view_by_id: dict[ViewEntity, DMSView],
    ) -> dict[tuple[ViewEntity, str], dm.DirectRelationReference]:
        edge_connection_property_by_view_property_id: dict[tuple[ViewEntity, str], DMSProperty] = {}
        for properties in view_properties_with_ancestors_by_id.values():
            for prop in properties:
                if isinstance(prop.connection, EdgeEntity):
                    view_property_id = (prop.view, prop.view_property)
                    edge_connection_property_by_view_property_id[view_property_id] = prop

        edge_types_by_view_property_id: dict[tuple[ViewEntity, str], dm.DirectRelationReference] = {}

        outwards_type_by_view_value_type: dict[tuple[ViewEntity, ViewEntity], list[dm.DirectRelationReference]] = (
            defaultdict(list)
        )
        # First set the edge types for outwards connections.
        for (view_id, _), prop in edge_connection_property_by_view_property_id.items():
            # We have already filtered out all non-EdgeEntity connections
            connection = cast(EdgeEntity, prop.connection)
            if connection.direction == "inwards":
                continue
            view = view_by_id[view_id]

            edge_type = cls._get_edge_type_outwards_connection(
                view, prop, view_by_id, edge_connection_property_by_view_property_id
            )

            edge_types_by_view_property_id[(prop.view, prop.view_property)] = edge_type

            if isinstance(prop.value_type, ViewEntity):
                outwards_type_by_view_value_type[(prop.value_type, prop.view)].append(edge_type)

        # Then inwards connections = outwards connections
        for (view_id, prop_id), prop in edge_connection_property_by_view_property_id.items():
            # We have already filtered out all non-EdgeEntity connections
            connection = cast(EdgeEntity, prop.connection)

            if connection.direction == "inwards" and isinstance(prop.value_type, ViewEntity):
                edge_type_candidates = outwards_type_by_view_value_type.get((prop.view, prop.value_type), [])
                if len(edge_type_candidates) == 0:
                    # Warning in validation, should not have an inwards connection without an outwards connection
                    edge_type = cls._default_edge_type_from_view_id(prop.view.as_id(), prop_id)
                elif len(edge_type_candidates) == 1:
                    edge_type = edge_type_candidates[0]
                else:
                    raise NeatValueError(
                        f"Cannot infer edge type for {view_id}.{prop_id}, multiple candidates: {edge_type_candidates}."
                        "Please specify edge type explicitly, i.e., edge(type=<YOUR_TYPE>)."
                    )
                view_property_id = (prop.view, prop.view_property)
                edge_types_by_view_property_id[view_property_id] = edge_type

        return edge_types_by_view_property_id

    @classmethod
    def _get_edge_type_outwards_connection(
        cls,
        view: DMSView,
        prop: DMSProperty,
        view_by_id: dict[ViewEntity, DMSView],
        edge_connection_by_view_property_id: dict[tuple[ViewEntity, str], DMSProperty],
    ) -> dm.DirectRelationReference:
        connection = cast(EdgeEntity, prop.connection)
        if connection.edge_type is not None:
            # Explicitly set edge type
            return connection.edge_type.as_reference()
        elif view.implements:
            # Try to look for same property in parent views
            candidates = []
            for parent_id in view.implements:
                if parent_view := view_by_id.get(parent_id):
                    parent_prop = edge_connection_by_view_property_id.get((parent_view.view, prop.view_property))
                    if parent_prop and isinstance(parent_prop.connection, EdgeEntity):
                        parent_edge_type = cls._get_edge_type_outwards_connection(
                            parent_view, parent_prop, view_by_id, edge_connection_by_view_property_id
                        )
                        candidates.append(parent_edge_type)
            if len(candidates) == 0:
                return cls._default_edge_type_from_view_id(prop.view.as_id(), prop.view_property)
            elif len(candidates) == 1:
                return candidates[0]
            else:
                raise NeatValueError(
                    f"Cannot infer edge type for {prop.view.as_id()!r}.{prop.view_property}, "
                    f"multiple candidates: {candidates}. "
                    "Please specify edge type explicitly, i.e., edge(type=<YOUR_TYPE>)."
                )
        else:
            # No parent view, use the default
            return cls._default_edge_type_from_view_id(prop.view.as_id(), prop.view_property)

    def _create_containers(
        self,
        container_properties_by_id: dict[dm.ContainerId, list[DMSProperty]],
        enum: Collection[DMSEnum] | None,
    ) -> ContainerApplyDict:
        enum_values_by_collection: dict[ClassEntity, list[DMSEnum]] = defaultdict(list)
        for enum_value in enum or []:
            enum_values_by_collection[enum_value.collection].append(enum_value)

        containers = list(self.rules.containers or [])

        containers = dm.ContainerApplyList(
            [
                dms_container.as_container()
                for dms_container in containers
                if not (self.remove_cdf_spaces and dms_container.container.space in COGNITE_SPACES)
            ]
        )
        container_to_drop = set()
        for container in containers:
            container_id = container.as_id()
            if not (container_properties := container_properties_by_id.get(container_id)):
                if container_id.space not in COGNITE_SPACES:
                    warnings.warn(
                        EmptyContainerWarning(container_id),
                        stacklevel=2,
                    )
                container_to_drop.add(container_id)
                continue
            for prop in container_properties:
                if prop.container_property is None:
                    continue
                if isinstance(prop.value_type, DataType):
                    type_cls = prop.value_type.dms
                else:
                    type_cls = dm.DirectRelation

                args: dict[str, Any] = {}
                if issubclass(type_cls, ListablePropertyType):
                    args["is_list"] = prop.is_list or False
                if isinstance(prop.value_type, Double | Float) and isinstance(prop.value_type.unit, UnitEntity):
                    args["unit"] = prop.value_type.unit.as_reference()
                if isinstance(prop.value_type, Enum):
                    if prop.value_type.collection not in enum_values_by_collection:
                        raise ResourceNotFoundError(
                            prop.value_type.collection, "enum collection", prop.container, "container"
                        )
                    args["unknown_value"] = prop.value_type.unknown_value
                    args["values"] = {
                        value.value: DMSEnumValue(
                            name=value.name,
                            description=value.description,
                        )
                        for value in enum_values_by_collection[prop.value_type.collection]
                    }

                type_ = type_cls(**args)
                container.properties[prop.container_property] = dm.ContainerProperty(
                    type=type_,
                    # If not set, nullable is True and immutable is False
                    nullable=prop.nullable if prop.nullable is not None else True,
                    immutable=prop.immutable if prop.immutable is not None else False,
                    # Guarding against default value being set for connection properties
                    default_value=prop.default if not prop.connection else None,
                    name=prop.name,
                    description=prop.description,
                )

            uniqueness_properties: dict[str, set[str]] = defaultdict(set)
            for prop in container_properties:
                if prop.container_property is not None:
                    for constraint in prop.constraint or []:
                        uniqueness_properties[constraint].add(prop.container_property)
            for constraint_name, properties in uniqueness_properties.items():
                container.constraints = container.constraints or {}
                container.constraints[constraint_name] = dm.UniquenessConstraint(properties=list(properties))

            index_properties: dict[str, set[str]] = defaultdict(set)
            for prop in container_properties:
                if prop.container_property is not None:
                    for index in prop.index or []:
                        index_properties[index].add(prop.container_property)
            for index_name, properties in index_properties.items():
                container.indexes = container.indexes or {}
                container.indexes[index_name] = BTreeIndex(properties=list(properties))

        # We might drop containers we convert direct relations of list into multi-edge connections
        # which do not have a container.
        for container in containers:
            if container.constraints:
                container.constraints = {
                    name: const
                    for name, const in container.constraints.items()
                    if not (isinstance(const, dm.RequiresConstraint) and const.require in container_to_drop)
                }
        return ContainerApplyDict([container for container in containers if container.as_id() not in container_to_drop])

    @staticmethod
    def _gather_properties(
        properties: Sequence[DMSProperty],
    ) -> tuple[dict[dm.ContainerId, list[DMSProperty]], dict[dm.ViewId, list[DMSProperty]]]:
        container_properties_by_id: dict[dm.ContainerId, list[DMSProperty]] = defaultdict(list)
        view_properties_by_id: dict[dm.ViewId, list[DMSProperty]] = defaultdict(list)
        for prop in properties:
            view_id = prop.view.as_id()
            view_properties_by_id[view_id].append(prop)

            if prop.container and prop.container_property:
                container_id = prop.container.as_id()
                container_properties_by_id[container_id].append(prop)

        return container_properties_by_id, view_properties_by_id

    def _gather_properties_with_ancestors(
        self,
        view_properties_by_id: dict[dm.ViewId, list[DMSProperty]],
        views: Sequence[DMSView],
    ) -> dict[dm.ViewId, list[DMSProperty]]:
        all_view_properties_by_id = view_properties_by_id.copy()

        view_properties_with_parents_by_id: dict[dm.ViewId, list[DMSProperty]] = defaultdict(list)
        view_by_view_id = {view.view.as_id(): view for view in views}
        for view in views:
            view_id = view.view.as_id()
            seen: set[Hashable] = set()
            if view_properties := all_view_properties_by_id.get(view_id):
                view_properties_with_parents_by_id[view_id].extend(view_properties)
                seen.update(prop._identifier() for prop in view_properties)
            if not view.implements:
                continue
            parents = view.implements.copy()
            seen_parents = set(parents)
            while parents:
                parent = parents.pop()
                parent_view_id = parent.as_id()
                if parent_view := view_by_view_id.get(parent_view_id):
                    for grandparent in parent_view.implements or []:
                        if grandparent not in seen_parents:
                            parents.append(grandparent)
                            seen_parents.add(grandparent)

                if not (parent_view_properties := all_view_properties_by_id.get(parent_view_id)):
                    continue
                for prop in parent_view_properties:
                    new_prop = prop.model_copy(update={"view": view.view})

                    if new_prop._identifier() not in seen:
                        view_properties_with_parents_by_id[view_id].append(new_prop)
                        seen.add(new_prop._identifier())

        return view_properties_with_parents_by_id

    @classmethod
    def _update_with_properties(
        cls,
        selected_properties: Sequence[DMSProperty],
        container_properties_by_id: dict[dm.ContainerId, list[DMSProperty]],
        view_properties_by_id: dict[dm.ViewId, list[DMSProperty]] | None,
        include_new_containers: bool = False,
    ) -> None:
        view_properties_by_id = view_properties_by_id or {}
        last_container_properties_by_id, last_view_properties_by_id = cls._gather_properties(selected_properties)

        for container_id, properties in last_container_properties_by_id.items():
            # Only add the container properties that are not already present, and do not overwrite.
            if (container_id in container_properties_by_id) or include_new_containers:
                existing = {prop.container_property for prop in container_properties_by_id.get(container_id, [])}
                container_properties_by_id[container_id].extend(
                    [prop for prop in properties if prop.container_property not in existing]
                )

        if view_properties_by_id:
            for view_id, properties in last_view_properties_by_id.items():
                existing = {prop.view_property for prop in view_properties_by_id[view_id]}
                view_properties_by_id[view_id].extend(
                    [prop for prop in properties if prop.view_property not in existing]
                )

    def _create_view_filter(
        self,
        view: dm.ViewApply,
        dms_view: DMSView | None,
    ) -> DMSFilter | None:
        selected_filter_name = (dms_view and dms_view.filter_ and dms_view.filter_.name) or ""

        if dms_view and dms_view.filter_ and not dms_view.filter_.is_empty:
            # Has Explicit Filter, use it
            return dms_view.filter_

        # Enterprise Model or (Solution + HasData)
        ref_containers = view.referenced_containers()
        if not ref_containers or selected_filter_name == HasDataFilter.name:
            # Child filter without container properties
            if selected_filter_name == HasDataFilter.name:
                warnings.warn(
                    HasDataFilterOnNoPropertiesViewWarning(view.as_id()),
                    stacklevel=2,
                )
            return NodeTypeFilter(inner=[DMSNodeEntity(space=view.space, externalId=view.external_id)])
        else:
            # HasData or not provided (this is the default)
            return HasDataFilter(inner=[ContainerEntity.from_id(id_) for id_ in ref_containers])

    @classmethod
    def _create_view_property(
        cls,
        prop: DMSProperty,
        view_properties_with_ancestors_by_id: dict[dm.ViewId, list[DMSProperty]],
        edge_types_by_view_property_id: dict[tuple[ViewEntity, str], dm.DirectRelationReference],
    ) -> ViewPropertyApply | None:
        if prop.container and prop.container_property:
            return cls._create_mapped_property(prop)
        elif isinstance(prop.connection, EdgeEntity):
            return cls._create_edge_property(prop, edge_types_by_view_property_id)
        elif isinstance(prop.connection, ReverseConnectionEntity):
            return cls._create_reverse_direct_relation(prop, view_properties_with_ancestors_by_id)
        elif prop.view and prop.view_property and prop.connection:
            warnings.warn(
                NotSupportedWarning(f"{prop.connection} in {prop.view.as_id()!r}.{prop.view_property}"), stacklevel=2
            )
        return None

    @classmethod
    def _create_mapped_property(cls, prop: DMSProperty) -> dm.MappedPropertyApply:
        container = cast(ContainerEntity, prop.container)
        container_prop_identifier = cast(str, prop.container_property)
        extra_args: dict[str, Any] = {}
        if prop.connection == "direct":
            if isinstance(prop.value_type, ViewEntity):
                extra_args["source"] = prop.value_type.as_id()
            elif isinstance(prop.value_type, DMSUnknownEntity):
                extra_args["source"] = None
            else:
                # Should have been validated.
                raise ValueError(
                    "If this error occurs it is a bug in NEAT, please report"
                    f"Debug Info, Invalid valueType direct: {prop.model_dump_json()}"
                )
        elif prop.connection is not None:
            # Should have been validated.
            raise ValueError(
                "If this error occurs it is a bug in NEAT, please report"
                f"Debug Info, Invalid connection: {prop.model_dump_json()}"
            )
        return dm.MappedPropertyApply(
            container=container.as_id(),
            container_property_identifier=container_prop_identifier,
            name=prop.name,
            description=prop.description,
            **extra_args,
        )

    @classmethod
    def _create_edge_property(
        cls, prop: DMSProperty, edge_types_by_view_property_id: dict[tuple[ViewEntity, str], dm.DirectRelationReference]
    ) -> dm.EdgeConnectionApply:
        connection = cast(EdgeEntity, prop.connection)
        if isinstance(prop.value_type, ViewEntity):
            source_view_id = prop.value_type.as_id()
        else:
            # Should have been validated.
            raise ValueError(
                "If this error occurs it is a bug in NEAT, please report"
                f"Debug Info, Invalid valueType edge: {prop.model_dump_json()}"
            )
        edge_source: dm.ViewId | None = None
        if connection.properties is not None:
            edge_source = connection.properties.as_id()
        edge_cls: type[dm.EdgeConnectionApply] = dm.MultiEdgeConnectionApply
        # If is_list is not set, we default to a MultiEdgeConnection
        if prop.is_list is False:
            edge_cls = SingleEdgeConnectionApply

        return edge_cls(
            type=edge_types_by_view_property_id[(prop.view, prop.view_property)],
            source=source_view_id,
            direction=connection.direction,
            name=prop.name,
            description=prop.description,
            edge_source=edge_source,
        )

    @classmethod
    def _create_reverse_direct_relation(
        cls, prop: DMSProperty, view_properties_with_ancestors_by_id: dict[dm.ViewId, list[DMSProperty]]
    ) -> dm.MultiReverseDirectRelationApply | SingleReverseDirectRelationApply | None:
        connection = cast(ReverseConnectionEntity, prop.connection)
        reverse_prop_id = connection.property_
        if isinstance(prop.value_type, ViewEntity):
            source_view_id = prop.value_type.as_id()
        else:
            # Should have been validated.
            raise ValueError(
                "If this error occurs it is a bug in NEAT, please report"
                f"Debug Info, Invalid valueType reverse connection: {prop.model_dump_json()}"
            )
        reverse_prop = next(
            (
                prop
                for prop in view_properties_with_ancestors_by_id.get(source_view_id, [])
                if prop.view_property == reverse_prop_id
            ),
            None,
        )
        if reverse_prop is None:
            warnings.warn(
                PropertyNotFoundWarning(
                    source_view_id,
                    "view",
                    reverse_prop_id or "MISSING",
                    dm.PropertyId(prop.view.as_id(), prop.view_property),
                    "view property",
                ),
                stacklevel=2,
            )

        if reverse_prop and reverse_prop.connection == "direct":
            args: dict[str, Any] = dict(
                source=source_view_id,
                through=dm.PropertyId(source=source_view_id, property=reverse_prop_id),
                name=prop.name,
                description=prop.description,
            )
            if prop.is_list in [True, None]:
                return dm.MultiReverseDirectRelationApply(**args)
            else:
                return SingleReverseDirectRelationApply(**args)
        else:
            return None



from typing import Literal

from cognite.neat._issues.errors import NeatTypeError
from cognite.neat._rules.models.data_types import DataType

from ._multi_value import MultiValueTypeInfo
from ._single_value import (
    ClassEntity,
    DMSUnknownEntity,
    EdgeEntity,
    ReverseConnectionEntity,
    Unknown,
    UnknownEntity,
    ViewEntity,
)


def load_value_type(
    raw: str | MultiValueTypeInfo | DataType | ClassEntity | UnknownEntity, default_prefix: str
) -> MultiValueTypeInfo | DataType | ClassEntity | UnknownEntity:
    if isinstance(raw, MultiValueTypeInfo | DataType | ClassEntity | UnknownEntity):
        return raw
    elif isinstance(raw, str):
        # property holding xsd data type
        # check if it is multi value type
        if "," in raw:
            value_type = MultiValueTypeInfo.load(raw)
            value_type.set_default_prefix(default_prefix)
            return value_type
        elif DataType.is_data_type(raw):
            return DataType.load(raw)

        # unknown value type
        elif raw == str(Unknown):
            return UnknownEntity()

        # property holding link to class
        else:
            return ClassEntity.load(raw, prefix=default_prefix)
    else:
        raise NeatTypeError(f"Invalid value type: {type(raw)}")


def load_dms_value_type(
    raw: str | DataType | ViewEntity | DMSUnknownEntity,
    default_space: str,
    default_version: str,
) -> DataType | ViewEntity | DMSUnknownEntity:
    if isinstance(raw, DataType | ViewEntity | DMSUnknownEntity):
        return raw
    elif isinstance(raw, str):
        if DataType.is_data_type(raw):
            return DataType.load(raw)
        elif raw == str(Unknown):
            return DMSUnknownEntity()
        else:
            return ViewEntity.load(raw, space=default_space, version=default_version)
    raise NeatTypeError(f"Invalid value type: {type(raw)}")


def load_connection(
    raw: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | str | None,
    default_space: str,
    default_version: str,
) -> Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None:
    if isinstance(raw, str) and raw.lower() == "direct":
        return "direct"  # type: ignore[return-value]
    elif isinstance(raw, EdgeEntity | ReverseConnectionEntity) or raw is None:
        return raw  # type: ignore[return-value]
    elif isinstance(raw, str) and raw.startswith("edge"):
        return EdgeEntity.load(raw, space=default_space, version=default_version)  # type: ignore[return-value]
    elif isinstance(raw, str) and raw.startswith("reverse"):
        return ReverseConnectionEntity.load(raw)  # type: ignore[return-value]
    raise NeatTypeError(f"Invalid connection: {type(raw)}")



from typing import Literal

from cognite.client.data_classes.data_modeling import DataModelIdentifier

from cognite.neat._issues import IssueList
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.models.dms import DMSValidation
from cognite.neat._rules.transformers import (
    IncludeReferenced,
    ToDataProductModel,
    ToEnterpriseModel,
    ToSolutionModel,
    VerifiedRulesTransformer,
)

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class CreateAPI:
    """
    Create new data model based on the given data.
    """

    def __init__(self, state: SessionState):
        self._state = state

    def enterprise_model(
        self,
        data_model_id: DataModelIdentifier,
        org_name: str = "My",
        dummy_property: str = "GUID",
    ) -> IssueList:
        """Uses the current data model as a basis to create enterprise data model

        Args:
            data_model_id: The enterprise data model id that is being created
            org_name: Organization name to use for the views in the enterprise data model.
            dummy_property: The dummy property to use as placeholder for the views in the new data model.

        What does this function do?
            1. It creates a new view for each view in the current data model that implements the view it is based on.
            2. If dummy_property is set, it will create a container with one property for each view and connect the
               view to the container.
            3. It will repeat all connection properties in the new views and update the ValueTypes to match the new
               views.

        !!! note "Enterprise Data Model Creation"

            Always create an enterprise data model from a Cognite Data Model as this will
            assure all the Cognite Data Fusion applications to run smoothly, such as
                - Search
                - Atlas AI
                - Infield
                - Canvas
                - Maintain
                - Charts

        """
        last_rules = self._get_last_rules()
        issues = self._state.rule_transform(
            ToEnterpriseModel(
                new_model_id=data_model_id,
                org_name=org_name,
                dummy_property=dummy_property,
                move_connections=True,
            )
        )
        if last_rules and not issues.has_errors:
            self._state.last_reference = last_rules
        return issues

    def _get_last_rules(self) -> InformationRules | DMSRules | None:
        if not self._state.rule_store.provenance:
            return None
        last_entity = self._state.rule_store.provenance[-1].target_entity
        return last_entity.dms or last_entity.information

    def solution_model(
        self,
        data_model_id: DataModelIdentifier,
        direct_property: str = "enterprise",
        view_prefix: str = "Enterprise",
    ) -> IssueList:
        """Uses the current data model as a basis to create solution data model

        Args:
            data_model_id: The solution data model id that is being created.
            direct_property: The property to use for the direct connection between the views in the solution data model
                and the enterprise data model.
            view_prefix: The prefix to use for the views in the enterprise data model.

        What does this function do?
        1. It will create two new views for each view in the current data model. The first view will be read-only and
           prefixed with the 'view_prefix'. The second view will be writable and have one property that connects to the
           read-only view named 'direct_property'.
        2. It will repeat all connection properties in the new views and update the ValueTypes to match the new views.
        3. Each writable view will have a container with the single property that connects to the read-only view.

        !!! note "Solution Data Model Mode"

            The read-only solution model will only be able to read from the existing containers
            from the enterprise data model, therefore the solution data model will not have
            containers in the solution data model space. Meaning the solution data model views
            will be read-only.

            The write mode will have additional containers in the solution data model space,
            allowing in addition to read through the solution model views, also writing to
            the containers in the solution data model space.

        """
        last_rules = self._get_last_rules()
        issues = self._state.rule_transform(
            ToSolutionModel(
                new_model_id=data_model_id,
                properties="connection",
                direct_property=direct_property,
                view_prefix=view_prefix,
            )
        )
        if last_rules and not issues.has_errors:
            self._state.last_reference = last_rules
        return issues

    def data_product_model(
        self,
        data_model_id: DataModelIdentifier,
        include: Literal["same-space", "all"] = "same-space",
    ) -> IssueList:
        """Uses the current data model as a basis to create data product data model.

        A data product model is a data model that ONLY maps to containers and do not use implements. This is
        typically used for defining the data in a data product.

        What does this function do?
        1. It creates a new view for each view in the current data model. The new views uses the same filter
           as the view it is based on.
        2. It will repeat all connection properties in the new views and update the ValueTypes to match the new views.

        Args:
            data_model_id: The data product data model id that is being created.
            include: The views to include in the data product data model. Can be either "same-space" or "all".
                If you set same-space, only the properties of the views in the same space as the data model
                will be included.
        """
        last_rules = self._get_last_rules()
        view_ids, container_ids = DMSValidation(
            self._state.rule_store.last_verified_dms_rules
        ).imported_views_and_containers_ids()
        transformers: list[VerifiedRulesTransformer] = []
        client = self._state.client
        if (view_ids or container_ids) and client is None:
            raise NeatSessionError(
                "No client provided. You are referencing unknown views and containers in your data model, "
                "NEAT needs a client to lookup the definitions. "
                "Please set the client in the session, NeatSession(client=client)."
            )
        elif (view_ids or container_ids) and client:
            transformers.append(IncludeReferenced(client, include_properties=True))

        transformers.append(ToDataProductModel(new_model_id=data_model_id, include=include))

        issues = self._state.rule_transform(*transformers)
        if last_rules and not issues.has_errors:
            self._state.last_reference = last_rules
        return issues



from pathlib import Path
from typing import Literal

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.neat import _version
from cognite.neat._client import NeatClient
from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import RegexViolationError
from cognite.neat._issues.errors._general import NeatImportError
from cognite.neat._rules import importers
from cognite.neat._rules.models import DMSRules
from cognite.neat._rules.models.information._rules import InformationRules
from cognite.neat._rules.transformers import (
    InformationToDMS,
    MergeDMSRules,
    MergeInformationRules,
    ToDMSCompliantEntities,
    VerifyInformationRules,
)
from cognite.neat._store._rules_store import RulesEntity
from cognite.neat._utils.auxiliary import local_import

from ._collector import _COLLECTOR, Collector
from ._create import CreateAPI
from ._drop import DropAPI
from ._explore import ExploreAPI
from ._fix import FixAPI
from ._inspect import InspectAPI
from ._mapping import MappingAPI
from ._prepare import PrepareAPI
from ._read import ReadAPI
from ._set import SetAPI
from ._show import ShowAPI
from ._state import SessionState
from ._subset import SubsetAPI
from ._to import ToAPI
from .engine import load_neat_engine
from .exceptions import session_class_wrapper


@session_class_wrapper
class NeatSession:
    """Creates a new NeatSession.

    This is the main entry point for using Neat. It provides access to the different APIs that can be used to read,
    write, and manipulate data and data models.

    Args:
        client: The CogniteClient to use for reading and writing data.
        storage: The storage type to use for storing data and data models. Can be either "memory" or "oxigraph".
            In "memory" mode works well for small data sets and when only working with data models. It is works
            well for all notebook environments. In "oxigraph" mode, the data is stored in an Oxigraph database. This
            is more performant for larger data sets and when working with data. Note that this option requires
            additional dependencies to be installed and is not available in CDF Notebooks.
        verbose: Whether to print information about the operations being performed.
        load_engine: Whether to load the Neat Engine. Can be "newest", "cache", or "skip". "newest" will always
            check for the newest version of the engine. "cache" will load the engine if it has been downloaded before.
            "skip" will not load the engine.

    Example:
        Instantiate a NeatSession outside CDF jupyter notebook (needs instantiation of a CogniteClient)
        ```python
        from cognite.neat import get_cognite_client
        from cognite.neat import NeatSession

        client = get_cognite_client(env_file_name=".env")
        neat = NeatSession(client)
        ```

    Example:
        Instantiate a NeatSession inside a CDF jupyter notebook (use your user's CogniteClient directly)
        ```python
        from cognite.client import CogniteClient
        from cognite.neat import NeatSession

        client = CogniteClient()
        neat = NeatSession(client)
        ```
    """

    def __init__(
        self,
        client: CogniteClient | None = None,
        storage: Literal["memory", "oxigraph"] | None = None,
        storage_path: str | None = None,
        verbose: bool = True,
        load_engine: Literal["newest", "cache", "skip"] = "cache",
    ) -> None:
        self._verbose = verbose
        self._state = SessionState(
            store_type=storage or self._select_most_performant_store(),
            storage_path=Path(storage_path) if storage_path else None,
            client=NeatClient(client) if client else None,
        )
        self.read = ReadAPI(self._state, verbose)
        self.to = ToAPI(self._state, verbose)
        self.fix = FixAPI(self._state, verbose)
        self.prepare = PrepareAPI(self._state, verbose)
        self.show = ShowAPI(self._state)
        self.set = SetAPI(self._state, verbose)
        self.inspect = InspectAPI(self._state)
        self.mapping = MappingAPI(self._state)
        self.drop = DropAPI(self._state)
        self.subset = SubsetAPI(self._state)
        self.create = CreateAPI(self._state)
        self._explore = ExploreAPI(self._state)
        self.opt = OptAPI()
        self.opt._display()
        if load_engine != "skip" and (engine_version := load_neat_engine(client, load_engine)):
            print(f"Neat Engine {engine_version} loaded.")

    def _select_most_performant_store(self) -> Literal["memory", "oxigraph"]:
        """Select the most performant store based on the current environment."""

        try:
            local_import("pyoxigraph", "oxi")
            local_import("oxrdflib", "oxi")
            return "oxigraph"
        except NeatImportError:
            return "memory"

    @property
    def version(self) -> str:
        """Get the current version of neat.

        Returns:
            The current version of neat used in the session.

        Example:
            ```python
            neat.version
            ```
        """
        return _version.__version__

    def verify(self) -> IssueList:
        """
        Verify the Data Model schema before the model can be written to CDF. If verification was unsuccessful, use
        `.inspect.issues()` to see what went wrong.

        Example:
            Verify a data model after reading a source file and inferring the data model
            ```python
            # From an active NeatSession
            ...
            neat.read.xml.dexpi("url_or_path_to_dexpi_file")
            neat.infer()
            neat.verify()
            ```
        """
        print("This action has no effect. Neat no longer supports unverified data models.")
        return IssueList()

    def convert(self, reserved_properties: Literal["error", "warning"] = "warning") -> IssueList:
        """Converts the last verified data model to the target type.

        Args:
            reserved_properties: What to do with reserved properties. Can be "error" or "warning".

        Example:
            Convert to DMS rules
            ```python
            neat.convert()
            ```
        """
        converter = InformationToDMS(reserved_properties=reserved_properties)

        issues = self._state.rule_transform(converter)

        if self._verbose and not issues.has_errors:
            print("Rules converted to dms.")
        else:
            print("Conversion failed.")
        if issues:
            print("You can inspect the issues with the .inspect.issues(...) method.")
            if issues.has_error_type(RegexViolationError):
                print("You can use .prepare. to try to fix the issues")

        return issues

    def infer(
        self,
        model_id: dm.DataModelId | tuple[str, str, str] = (
            "neat_space",
            "NeatInferredDataModel",
            "v1",
        ),
    ) -> IssueList:
        """Data model inference from instances.

        Args:
            model_id: The ID of the inferred data model.

        Example:
            Infer a data model after reading a source file
            ```python
            # From an active NeatSession
            ...
            neat.read.xml.dexpi("url_or_path_to_dexpi_file")
            neat.infer()
            ```
        """
        return self._infer_subclasses(model_id)

    def _previous_inference(
        self, model_id: dm.DataModelId | tuple[str, str, str], max_number_of_instance: int = 100
    ) -> IssueList:
        # Temporary keeping the old inference method in case we need to revert back
        model_id = dm.DataModelId.load(model_id)
        importer = importers.InferenceImporter.from_graph_store(
            store=self._state.instances.store,
            max_number_of_instance=max_number_of_instance,
            data_model_id=model_id,
        )
        return self._state.rule_import(importer)

    def _infer_subclasses(
        self,
        model_id: dm.DataModelId | tuple[str, str, str] = (
            "neat_space",
            "NeatInferredDataModel",
            "v1",
        ),
    ) -> IssueList:
        """Infer data model from instances."""
        last_entity: RulesEntity | None = None
        if self._state.rule_store.provenance:
            last_entity = self._state.rule_store.provenance[-1].target_entity

        # Note that this importer behaves as a transformer in the rule store when there is an existing rules.
        # We are essentially transforming the last entity's information rules into a new set of information rules.
        importer = importers.SubclassInferenceImporter(
            issue_list=IssueList(),
            graph=self._state.instances.store.graph(),
            rules=last_entity.information if last_entity is not None else None,
            data_model_id=dm.DataModelId.load(model_id) if last_entity is None else None,
        )

        def action() -> tuple[InformationRules, DMSRules | None]:
            unverified_information = importer.to_rules()
            unverified_information = ToDMSCompliantEntities(rename_warning="raise").transform(unverified_information)

            extra_info = VerifyInformationRules().transform(unverified_information)
            if not last_entity:
                return extra_info, None
            merged_info = MergeInformationRules(extra_info).transform(last_entity.information)
            if not last_entity.dms:
                return merged_info, None

            extra_dms = InformationToDMS(reserved_properties="warning").transform(extra_info)

            merged_dms = MergeDMSRules(extra_dms).transform(last_entity.dms)
            return merged_info, merged_dms

        return self._state.rule_store.do_activity(action, importer)

    def _repr_html_(self) -> str:
        state = self._state
        if state.instances.empty and state.rule_store.empty:
            return "<strong>Empty session</strong>. Get started by reading something with the <em>.read</em> attribute."

        output = []

        if state.rule_store.provenance:
            last_entity = state.rule_store.provenance[-1].target_entity
            if last_entity.dms:
                html = last_entity.dms._repr_html_()
            else:
                html = last_entity.information._repr_html_()
            output.append(f"<H2>Data Model</H2><br />{html}")  # type: ignore

        if not state.instances.empty:
            output.append(f"<H2>Instances</H2> {state.instances.store._repr_html_()}")

        return "<br />".join(output)

    def close(self) -> None:
        """Close the session and release resources."""
        self._state.instances.store.dataset.close()

    def __del__(self) -> None:
        """Called by garbage collector"""
        self.close()


@session_class_wrapper
class OptAPI:
    """For the user to decide if they want their usage of neat to be collected or not. We do not collect personal
    information like name etc. only usage.
    """

    def __init__(self, collector: Collector | None = None) -> None:
        self._collector = collector or _COLLECTOR

    def _display(self) -> None:
        if self._collector.opted_in or self._collector.opted_out:
            return
        print(
            "For Neat to improve, we need to collect usage information. "
            "You acknowledge and agree that neat may collect usage information."
            "To remove this message run 'neat.opt.in_() "
            "or to stop collecting usage information run 'neat.opt.out()'."
        )

    def in_(self) -> None:
        """Consent to collection of neat user insights."""
        self._collector.enable()
        print("You have successfully opted in to data collection.")

    def out(self) -> None:
        """Opt out of allowing usage of neat to be collected from current user."""
        self._collector.disable()
        print("You have successfully opted out of data collection.")



from dataclasses import dataclass
from pathlib import Path

from cognite.neat._issues import NeatWarning


@dataclass(unsafe_hash=True)
class FileReadWarning(NeatWarning):
    """Error when reading file, {filepath}: {reason}"""

    filepath: Path
    reason: str


@dataclass(unsafe_hash=True)
class FileMissingRequiredFieldWarning(NeatWarning):
    """Missing required {field_name} in {filepath}: {field}. The file will be skipped"""

    filepath: Path
    field_name: str
    field: str


@dataclass(unsafe_hash=True)
class FileTypeUnexpectedWarning(NeatWarning):
    """Unexpected file type: {filepath}. Expected format: {expected_format}"""

    extra = "Error: {error_message}"

    filepath: Path
    expected_format: frozenset[str]
    error_message: str | None = None


@dataclass(unsafe_hash=True)
class FileItemNotSupportedWarning(NeatWarning):
    """The item {item} in {filepath} is not supported. The item will be skipped"""

    item: str
    filepath: Path


@dataclass(unsafe_hash=True)
class CDFAuthWarning(NeatWarning):
    """Failed to {action}: {reason}"""

    action: str
    reason: str


@dataclass(unsafe_hash=True)
class CDFMaxIterationsWarning(NeatWarning):
    """The maximum number of iterations ({max_iterations}) has been reached. {message}"""

    message: str
    max_iterations: int



import warnings
from collections import Counter, defaultdict
from collections.abc import Mapping
from functools import lru_cache
from typing import ClassVar

from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import ContainerList, ViewId, ViewList
from cognite.client.data_classes.data_modeling.views import (
    ReverseDirectRelation,
    ReverseDirectRelationApply,
    ViewProperty,
    ViewPropertyApply,
)

from cognite.neat._client import NeatClient
from cognite.neat._client.data_classes.data_modeling import ViewApplyDict
from cognite.neat._client.data_classes.schema import DMSSchema
from cognite.neat._constants import COGNITE_MODELS, DMS_CONTAINER_PROPERTY_SIZE_LIMIT, DMS_VIEW_CONTAINER_SIZE_LIMIT
from cognite.neat._issues import IssueList, NeatError
from cognite.neat._issues.errors import (
    CDFMissingClientError,
    PropertyDefinitionDuplicatedError,
    PropertyMappingDuplicatedError,
    PropertyNotFoundError,
    ResourceDuplicatedError,
    ResourceNotFoundError,
    ReversedConnectionNotFeasibleError,
)
from cognite.neat._issues.errors._external import CDFMissingResourcesError
from cognite.neat._issues.warnings import (
    NotSupportedHasDataFilterLimitWarning,
    NotSupportedViewContainerLimitWarning,
    UndefinedViewWarning,
    user_modeling,
)
from cognite.neat._issues.warnings.user_modeling import (
    ContainerPropertyLimitWarning,
    DirectRelationMissingSourceWarning,
    NotNeatSupportedFilterWarning,
)
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import ContainerEntity, RawFilter
from cognite.neat._rules.models.entities._single_value import (
    ViewEntity,
)
from cognite.neat._utils.spreadsheet import SpreadsheetRead
from cognite.neat._utils.text import humanize_collection

from ._rules import DMSProperty, DMSRules


class DMSValidation:
    """This class does all the validation of the DMS rules that have dependencies between
    components."""

    # When checking for changes extension=addition, we need to check if the new view has changed.
    # For example, changing the filter is allowed, but changing the properties is not.
    changeable_view_attributes: ClassVar[set[str]] = {"filter"}

    def __init__(
        self,
        rules: DMSRules,
        client: NeatClient | None = None,
        read_info_by_spreadsheet: dict[str, SpreadsheetRead] | None = None,
    ) -> None:
        self._rules = rules
        self._client = client
        self._metadata = rules.metadata
        self._properties = rules.properties
        self._containers = rules.containers
        self._views = rules.views
        self._read_info_by_spreadsheet = read_info_by_spreadsheet or {}

    def imported_views_and_containers_ids(
        self, include_views_with_no_properties: bool = True
    ) -> tuple[set[ViewEntity], set[ContainerEntity]]:
        existing_views = {view.view for view in self._views}
        imported_views: set[ViewEntity] = set()
        for view in self._views:
            for parent in view.implements or []:
                if parent not in existing_views:
                    imported_views.add(parent)
        existing_containers = {container.container for container in self._containers or []}
        imported_containers: set[ContainerEntity] = set()
        view_with_properties: set[ViewEntity] = set()
        for prop in self._properties:
            if prop.container and prop.container not in existing_containers:
                imported_containers.add(prop.container)
            if prop.view not in existing_views:
                imported_views.add(prop.view)
            view_with_properties.add(prop.view)

        if include_views_with_no_properties:
            extra_views = existing_views - view_with_properties
            imported_views.update({view for view in extra_views})

        return imported_views, imported_containers

    def validate(self) -> IssueList:
        imported_views, imported_containers = self.imported_views_and_containers_ids(
            include_views_with_no_properties=False
        )
        if (imported_views or imported_containers) and self._client is None:
            raise CDFMissingClientError(
                f"{self._rules.metadata.as_data_model_id()} has imported views and/or container: "
                f"{imported_views}, {imported_containers}."
            )
        referenced_views = ViewList([])
        referenced_containers = ContainerList([])
        if self._client:
            referenced_views = self._client.loaders.views.retrieve(
                list(imported_views), include_connected=True, include_ancestor=True
            )
            referenced_containers = self._client.loaders.containers.retrieve(
                list(imported_containers), include_connected=True
            )

            missing_views = {view.as_id() for view in imported_views} - {view.as_id() for view in referenced_views}
            missing_containers = {container.as_id() for container in imported_containers} - {
                container.as_id() for container in referenced_containers
            }

            if missing_views or missing_containers:
                raise CDFMissingResourcesError(resources=f"{missing_views.union(missing_containers)}")

        # Setup data structures for validation
        dms_schema = self._rules.as_schema()
        ref_view_by_id = {view.as_id(): view for view in referenced_views}
        ref_container_by_id = {container.as_id(): container for container in referenced_containers}
        # All containers and views are the Containers/Views in the DMSRules + the referenced ones
        all_containers_by_id: dict[dm.ContainerId, dm.ContainerApply | dm.Container] = {
            **dict(dms_schema.containers.items()),
            **ref_container_by_id,
        }
        all_views_by_id: dict[dm.ViewId, dm.ViewApply | dm.View] = {**dict(dms_schema.views.items()), **ref_view_by_id}
        properties_by_ids = self._as_properties_by_ids(dms_schema.views, ref_view_by_id)
        ref_properties_by_ids = self._as_properties_by_ids(ref_view_by_id, {})
        all_properties_by_ids = {**ref_properties_by_ids, **properties_by_ids}
        view_properties_by_id = self._as_view_properties_by_id(properties_by_ids)
        parents_view_ids_by_child_id = self._parent_view_ids_by_child_id(all_views_by_id)

        issue_list = IssueList()

        # Validated for duplicated resource
        issue_list.extend(self._duplicated_resources())

        # Neat DMS classes Validation
        # These are errors that can only happen due to the format of the Neat DMS classes
        issue_list.extend(self._validate_raw_filter())
        issue_list.extend(self._consistent_container_properties())
        issue_list.extend(self._validate_value_type_existence())
        issue_list.extend(
            self._validate_property_referenced_views_and_containers_exists(all_views_by_id, all_containers_by_id)
        )

        # SDK classes validation
        issue_list.extend(self._containers_are_proper_size(dms_schema))
        issue_list.extend(
            self._validate_reverse_connections(
                properties_by_ids, all_containers_by_id, parents_view_ids_by_child_id, all_properties_by_ids
            )
        )
        issue_list.extend(self._validate_schema(dms_schema, all_views_by_id, all_containers_by_id))
        issue_list.extend(self._validate_referenced_container_limits(dms_schema.views, view_properties_by_id))
        issue_list.extend(self._same_space_views_and_data_model())
        return issue_list

    def _same_space_views_and_data_model(self) -> IssueList:
        issue_list = IssueList()

        schema = self._rules.as_schema(remove_cdf_spaces=True)

        if schema.data_model and schema.views:
            data_model_space = schema.data_model.space
            views_spaces = {view.space for view in schema.views.values()}

            if data_model_space not in views_spaces:
                issue_list.append(
                    user_modeling.ViewsAndDataModelNotInSameSpaceWarning(
                        data_model_space=data_model_space,
                        views_spaces=humanize_collection(views_spaces),
                    )
                )

        return issue_list

    def _duplicated_resources(self) -> IssueList:
        issue_list = IssueList()

        properties_sheet = self._read_info_by_spreadsheet.get("Properties")
        views_sheet = self._read_info_by_spreadsheet.get("Views")
        containers_sheet = self._read_info_by_spreadsheet.get("Containers")

        visited = defaultdict(list)
        for row_no, property_ in enumerate(self._properties):
            visited[property_._identifier()].append(
                properties_sheet.adjusted_row_number(row_no) if properties_sheet else row_no + 1
            )

        for identifier, rows in visited.items():
            if len(rows) == 1:
                continue
            issue_list.append(
                ResourceDuplicatedError(
                    identifier[1],
                    "property",
                    (
                        f"the Properties sheet at row {humanize_collection(rows)} "
                        "if data model is read from a spreadsheet."
                    ),
                )
            )

        visited = defaultdict(list)
        for row_no, view in enumerate(self._views):
            visited[view._identifier()].append(views_sheet.adjusted_row_number(row_no) if views_sheet else row_no + 1)

        for identifier, rows in visited.items():
            if len(rows) == 1:
                continue
            issue_list.append(
                ResourceDuplicatedError(
                    identifier[0],
                    "view",
                    (f"the Views sheet at row {humanize_collection(rows)} if data model is read from a spreadsheet."),
                )
            )

        if self._containers:
            visited = defaultdict(list)
            for row_no, container in enumerate(self._containers):
                visited[container._identifier()].append(
                    containers_sheet.adjusted_row_number(row_no) if containers_sheet else row_no + 1
                )

            for identifier, rows in visited.items():
                if len(rows) == 1:
                    continue
                issue_list.append(
                    ResourceDuplicatedError(
                        identifier[0],
                        "container",
                        (
                            f"the Containers sheet at row {humanize_collection(rows)} "
                            "if data model is read from a spreadsheet."
                        ),
                    )
                )

        return issue_list

    @staticmethod
    def _as_properties_by_ids(
        view_by_id: Mapping[dm.ViewId, dm.ViewApply] | Mapping[dm.ViewId, dm.View],
        ref_view_by_id: dict[dm.ViewId, dm.View],
    ) -> dict[tuple[ViewId, str], ViewPropertyApply | ViewProperty]:
        # Priority DMS schema properties.
        # No need to do long lookups in ref_views as these already contain all ancestor properties.
        properties_by_id: dict[tuple[ViewId, str], ViewPropertyApply | ViewProperty] = {}
        for view_id, view in view_by_id.items():
            for prop_id, prop in (view.properties or {}).items():
                properties_by_id[(view_id, prop_id)] = prop
            if view.implements:
                to_check = view.implements.copy()
                while to_check:
                    parent_id = to_check.pop()
                    if parent_id in view_by_id:
                        # Priority of the DMS schema properties
                        parent_view = view_by_id[parent_id]
                        for prop_id, prop in (parent_view.properties or {}).items():
                            if (view_id, prop_id) not in properties_by_id:
                                properties_by_id[(view_id, prop_id)] = prop
                        to_check.extend(parent_view.implements or [])
                    elif parent_id in ref_view_by_id:
                        # SDK properties
                        parent_read_view = ref_view_by_id[parent_id]
                        for prop_id, read_prop in parent_read_view.properties.items():
                            if (view_id, prop_id) not in properties_by_id:
                                properties_by_id[(view_id, prop_id)] = read_prop
                        # Read format of views already includes all ancestor properties
                        # so no need to check further
                    else:
                        # Missing views are caught else where
                        continue

        return properties_by_id

    @staticmethod
    def _as_view_properties_by_id(
        properties_by_ids: dict[tuple[ViewId, str], ViewPropertyApply | ViewProperty],
    ) -> dict[ViewId, list[tuple[str, ViewProperty | ViewPropertyApply]]]:
        view_properties_by_id: dict[dm.ViewId, list[tuple[str, ViewProperty | ViewPropertyApply]]] = defaultdict(list)
        for (view_id, prop_id), prop in properties_by_ids.items():
            view_properties_by_id[view_id].append((prop_id, prop))
        return view_properties_by_id

    @staticmethod
    def _parent_view_ids_by_child_id(
        all_views_by_id: dict[dm.ViewId, dm.ViewApply | dm.View],
    ) -> dict[ViewId, set[ViewId]]:
        @lru_cache
        def get_parents(child_view_id: ViewId) -> set[ViewId]:
            if child_view_id not in all_views_by_id:
                return set()
            child_view = all_views_by_id[child_view_id]
            parents = set(child_view.implements or [])
            for parent_id in child_view.implements or []:
                parents.update(get_parents(parent_id))
            return parents

        parents_by_view: dict[dm.ViewId, set[dm.ViewId]] = {}
        for view_id in all_views_by_id:
            parents_by_view[view_id] = get_parents(view_id)
        return parents_by_view

    def _consistent_container_properties(self) -> IssueList:
        container_properties_by_id: dict[tuple[ContainerEntity, str], list[tuple[int, DMSProperty]]] = defaultdict(list)
        for prop_no, prop in enumerate(self._properties):
            if prop.container and prop.container_property:
                container_properties_by_id[(prop.container, prop.container_property)].append((prop_no, prop))
        properties_sheet = self._read_info_by_spreadsheet.get("Properties")
        errors = IssueList()
        for (container, prop_name), properties in container_properties_by_id.items():
            if len(properties) == 1:
                continue
            container_id = container.as_id()

            row_numbers = {prop_no for prop_no, _ in properties}
            if properties_sheet:
                row_numbers = {properties_sheet.adjusted_row_number(row_no) for row_no in row_numbers}
            value_types = {prop.value_type for _, prop in properties if prop.value_type}
            # The container type 'direct' is an exception. On a container the type direct can point to any
            # node. The value type is typically set on the view.
            is_all_direct = all(prop.connection == "direct" for _, prop in properties)
            if len(value_types) > 1 and not is_all_direct:
                errors.append(
                    PropertyDefinitionDuplicatedError[dm.ContainerId](
                        container_id,
                        "container",
                        prop_name,
                        frozenset({v.dms._type if isinstance(v, DataType) else str(v) for v in value_types}),
                        tuple(row_numbers),
                        "rows",
                    )
                )
            list_definitions = {prop.is_list for _, prop in properties if prop.is_list is not None}
            if len(list_definitions) > 1:
                errors.append(
                    PropertyDefinitionDuplicatedError[dm.ContainerId](
                        container_id,
                        "container",
                        prop_name,
                        frozenset(list_definitions),
                        tuple(row_numbers),
                        "rows",
                    )
                )
            nullable_definitions = {prop.nullable for _, prop in properties if prop.nullable is not None}
            if len(nullable_definitions) > 1:
                errors.append(
                    PropertyDefinitionDuplicatedError[dm.ContainerId](
                        container_id,
                        "container",
                        prop_name,
                        frozenset(nullable_definitions),
                        tuple(row_numbers),
                        "rows",
                    )
                )
            default_definitions = {prop.default for _, prop in properties if prop.default is not None}
            if len(default_definitions) > 1:
                errors.append(
                    PropertyDefinitionDuplicatedError[dm.ContainerId](
                        container_id,
                        "container",
                        prop_name,
                        frozenset(
                            tuple(f"{k}:{v}" for k, v in def_.items()) if isinstance(def_, dict) else def_
                            for def_ in default_definitions
                        ),
                        tuple(row_numbers),
                        "rows",
                    )
                )
            index_definitions = {",".join(prop.index) for _, prop in properties if prop.index is not None}
            if len(index_definitions) > 1:
                errors.append(
                    PropertyDefinitionDuplicatedError[dm.ContainerId](
                        container_id,
                        "container",
                        prop_name,
                        frozenset(index_definitions),
                        tuple(row_numbers),
                        "rows",
                    )
                )
            constraint_definitions = {
                ",".join(prop.constraint) for _, prop in properties if prop.constraint is not None
            }
            if len(constraint_definitions) > 1:
                errors.append(
                    PropertyDefinitionDuplicatedError[dm.ContainerId](
                        container_id,
                        "container",
                        prop_name,
                        frozenset(constraint_definitions),
                        tuple(row_numbers),
                        "rows",
                    )
                )

        return errors

    @staticmethod
    def _containers_are_proper_size(dms_schema: DMSSchema) -> IssueList:
        errors = IssueList()
        for container_id, container in dms_schema.containers.items():
            count = len(container.properties or {})
            if count > DMS_CONTAINER_PROPERTY_SIZE_LIMIT:
                errors.append(ContainerPropertyLimitWarning(container_id, count))

        return errors

    @staticmethod
    def _validate_referenced_container_limits(
        views: ViewApplyDict, view_properties_by_id: dict[dm.ViewId, list[tuple[str, ViewProperty | ViewPropertyApply]]]
    ) -> IssueList:
        issue_list = IssueList()
        for view_id, view in views.items():
            view_properties = view_properties_by_id.get(view_id, [])
            mapped_containers = {
                prop.container
                for _, prop in view_properties
                if isinstance(prop, dm.MappedPropertyApply | dm.MappedProperty)
            }

            if mapped_containers and len(mapped_containers) > DMS_VIEW_CONTAINER_SIZE_LIMIT:
                issue_list.append(
                    NotSupportedViewContainerLimitWarning(
                        view_id,
                        len(mapped_containers),
                    )
                )

            if view.filter and isinstance(view.filter, dm.filters.HasData) and len(view.filter.dump()["hasData"]) > 10:
                issue_list.append(
                    NotSupportedHasDataFilterLimitWarning(
                        view_id,
                        len(view.filter.dump()["hasData"]),
                    )
                )
        return issue_list

    def _validate_raw_filter(self) -> IssueList:
        issue_list = IssueList()
        for view in self._views:
            if view.filter_ and isinstance(view.filter_, RawFilter):
                issue_list.append(
                    NotNeatSupportedFilterWarning(view.view.as_id()),
                )
        return issue_list

    def _validate_value_type_existence(self) -> IssueList:
        views = {prop_.view for prop_ in self._properties}.union({view_.view for view_ in self._views})
        issue_list = IssueList()
        for prop_ in self._properties:
            if isinstance(prop_.value_type, ViewEntity) and prop_.value_type not in views:
                issue_list.append(
                    UndefinedViewWarning(
                        str(prop_.view),
                        str(prop_.value_type),
                        prop_.view_property,
                    )
                )
        return issue_list

    def _validate_property_referenced_views_and_containers_exists(
        self,
        view_by_id: dict[dm.ViewId, dm.ViewApply | dm.View],
        containers_by_id: dict[dm.ContainerId, dm.ContainerApply | dm.Container],
    ) -> IssueList:
        issue_list = IssueList()
        for prop in self._properties:
            if prop.container:
                container_id = prop.container.as_id()
                if container_id not in containers_by_id:
                    issue_list.append(
                        ResourceNotFoundError(
                            container_id,
                            "container",
                            prop.view,
                            "view",
                        )
                    )
                elif (
                    prop.container_property and prop.container_property not in containers_by_id[container_id].properties
                ):
                    issue_list.append(
                        PropertyNotFoundError(
                            prop.container,
                            "container property",
                            prop.container_property,
                            dm.PropertyId(prop.view.as_id(), prop.view_property),
                            "view property",
                        )
                    )

            if prop.view.as_id() not in view_by_id:
                issue_list.append(
                    ResourceNotFoundError(
                        prop.view,
                        "view",
                        prop.view_property,
                        "property",
                    )
                )

        return issue_list

    def _validate_reverse_connections(
        self,
        view_property_by_property_id: dict[tuple[dm.ViewId, str], ViewPropertyApply | ViewProperty],
        containers_by_id: dict[dm.ContainerId, dm.ContainerApply | dm.Container],
        parents_by_view: dict[dm.ViewId, set[dm.ViewId]],
        all_view_property_by_property_id: dict[tuple[dm.ViewId, str], ViewPropertyApply | ViewProperty],
    ) -> IssueList:
        issue_list = IssueList()
        # do not check for reverse connections in Cognite models
        if self._metadata.as_data_model_id() in COGNITE_MODELS:
            return issue_list

        for (view_id, prop_id), prop_ in view_property_by_property_id.items():
            if not isinstance(prop_, ReverseDirectRelationApply | ReverseDirectRelation):
                continue
            target_id = prop_.through.source, prop_.through.property
            if target_id not in all_view_property_by_property_id:
                issue_list.append(
                    ReversedConnectionNotFeasibleError(
                        view_id,
                        "reversed connection",
                        prop_id,
                        f"The {prop_.through.source} {prop_.through.property} does not exist",
                    )
                )
                continue
            if isinstance(target_id[0], dm.ContainerId):
                # Todo: How to handle this case? Should not happen if you created the model with Neat
                continue

            target_property = all_view_property_by_property_id[(target_id[0], target_id[1])]
            # Validate that the target is a direct relation pointing to the view_id
            is_direct_relation = False
            if isinstance(target_property, dm.MappedProperty) and isinstance(target_property.type, dm.DirectRelation):
                is_direct_relation = True
            elif isinstance(target_property, dm.MappedPropertyApply):
                container = containers_by_id[target_property.container]
                if target_property.container_property_identifier in container.properties:
                    container_property = container.properties[target_property.container_property_identifier]
                    if isinstance(container_property.type, dm.DirectRelation):
                        is_direct_relation = True
            if not is_direct_relation:
                issue_list.append(
                    ReversedConnectionNotFeasibleError(
                        view_id,
                        "reversed connection",
                        prop_id,
                        f"{prop_.through.source} {prop_.through.property} is not a direct relation",
                    )
                )
                continue
            if not (
                isinstance(target_property, dm.MappedPropertyApply | dm.MappedProperty)
                # The direct relation is pointing to the view_id or one of its parents
                and (target_property.source == view_id or target_property.source in parents_by_view[view_id])
            ):
                issue_list.append(
                    ReversedConnectionNotFeasibleError(
                        view_id,
                        "reversed connection",
                        prop_id,
                        f"{prop_.through.source} {prop_.through.property} is not pointing to {view_id}",
                    )
                )
        return issue_list

    @staticmethod
    def _validate_schema(
        schema: DMSSchema,
        view_by_id: dict[dm.ViewId, dm.ViewApply | dm.View],
        containers_by_id: dict[dm.ContainerId, dm.ContainerApply | dm.Container],
    ) -> IssueList:
        errors: set[NeatError] = set()
        defined_spaces = schema.spaces.copy()

        for container_id, container in schema.containers.items():
            if container.space not in defined_spaces:
                errors.add(ResourceNotFoundError(container.space, "space", container_id, "container"))
            for constraint in container.constraints.values():
                if isinstance(constraint, dm.RequiresConstraint) and constraint.require not in containers_by_id:
                    errors.add(ResourceNotFoundError(constraint.require, "container", container_id, "container"))

        for view_id, view in schema.views.items():
            if view.space not in defined_spaces:
                errors.add(ResourceNotFoundError(view.space, "space", view_id, "view"))

            for parent in view.implements or []:
                if parent not in view_by_id:
                    errors.add(PropertyNotFoundError(parent, "view", "implements", view_id, "view"))

            for prop_name, prop in (view.properties or {}).items():
                if isinstance(prop, dm.MappedPropertyApply):
                    ref_container = containers_by_id.get(prop.container)
                    if ref_container is None:
                        errors.add(ResourceNotFoundError(prop.container, "container", view_id, "view"))
                    elif prop.container_property_identifier not in ref_container.properties:
                        errors.add(
                            PropertyNotFoundError(
                                prop.container,
                                "container",
                                prop.container_property_identifier,
                                view_id,
                                "view",
                            )
                        )
                    else:
                        container_property = ref_container.properties[prop.container_property_identifier]

                        if isinstance(container_property.type, dm.DirectRelation) and prop.source is None:
                            warnings.warn(
                                DirectRelationMissingSourceWarning(view_id, prop_name),
                                stacklevel=2,
                            )

                if (
                    isinstance(prop, dm.EdgeConnectionApply | ReverseDirectRelationApply)
                    and prop.source not in view_by_id
                ):
                    errors.add(PropertyNotFoundError(prop.source, "view", prop_name, view_id, "view"))

                if (
                    isinstance(prop, dm.EdgeConnectionApply)
                    and prop.edge_source is not None
                    and prop.edge_source not in view_by_id
                ):
                    errors.add(PropertyNotFoundError(prop.edge_source, "view", prop_name, view_id, "view"))

            # This allows for multiple view properties to be mapped to the same container property,
            # as long as they have different external_id, otherwise this will lead to raising
            # error ContainerPropertyUsedMultipleTimesError
            property_count = Counter(
                (prop.container, prop.container_property_identifier, view_property_identifier)
                for view_property_identifier, prop in (view.properties or {}).items()
                if isinstance(prop, dm.MappedPropertyApply)
            )

            for (
                container_id,
                container_property_identifier,
                _,
            ), count in property_count.items():
                if count > 1:
                    view_properties = [
                        prop_name
                        for prop_name, prop in (view.properties or {}).items()
                        if isinstance(prop, dm.MappedPropertyApply)
                        and (prop.container, prop.container_property_identifier)
                        == (container_id, container_property_identifier)
                    ]
                    errors.add(
                        PropertyMappingDuplicatedError(
                            container_id,
                            "container",
                            container_property_identifier,
                            frozenset({dm.PropertyId(view_id, prop_name) for prop_name in view_properties}),
                            "view property",
                        )
                    )

        if schema.data_model:
            model = schema.data_model
            if model.space not in defined_spaces:
                errors.add(ResourceNotFoundError(model.space, "space", model.as_id(), "data model"))

            view_counts: dict[dm.ViewId, int] = defaultdict(int)
            for view_id_or_class in model.views or []:
                view_id = view_id_or_class if isinstance(view_id_or_class, dm.ViewId) else view_id_or_class.as_id()
                if view_id not in view_by_id:
                    errors.add(ResourceNotFoundError(view_id, "view", model.as_id(), "data model"))
                view_counts[view_id] += 1

            for view_id, count in view_counts.items():
                if count > 1:
                    errors.add(
                        ResourceDuplicatedError(
                            view_id,
                            "view",
                            f"DMS {model.as_id()!r}",
                        )
                    )

        return IssueList(list(errors))



from dataclasses import dataclass
from typing import Generic, TypeAlias, TypeVar

from cognite.neat._rules.models import (
    DMSRules,
    InformationRules,
)
from cognite.neat._rules.models.dms._rules_input import DMSInputRules
from cognite.neat._rules.models.information._rules_input import InformationInputRules
from cognite.neat._utils.spreadsheet import SpreadsheetRead

VerifiedRules: TypeAlias = InformationRules | DMSRules

T_VerifiedRules = TypeVar("T_VerifiedRules", bound=VerifiedRules)
InputRules: TypeAlias = DMSInputRules | InformationInputRules
T_InputRules = TypeVar("T_InputRules", bound=InputRules)


@dataclass
class ReadRules(Generic[T_InputRules]):
    """This represents a rules that has been read."""

    rules: T_InputRules | None
    read_context: dict[str, SpreadsheetRead]

    @classmethod
    def display_type_name(cls) -> str:
        return "UnverifiedModel"

    @property
    def display_name(self):
        if self.rules is None:
            return "FailedRead"
        return self.rules.display_name


ReadInputRules: TypeAlias = ReadRules[DMSInputRules] | ReadRules[InformationInputRules]
T_ReadInputRules = TypeVar("T_ReadInputRules", bound=ReadInputRules)

Rules: TypeAlias = InformationRules | DMSRules | ReadRules[DMSInputRules] | ReadRules[InformationInputRules]
T_Rules = TypeVar("T_Rules", bound=Rules)



from rdflib import Namespace

from cognite.neat._graph.extractors import IODDExtractor

from ._prune_graph import AttachPropertyFromTargetToSource, PruneDanglingNodes

IODD = Namespace("http://www.io-link.com/IODD/2010/10/")


class IODDAttachPropertyFromTargetToSource(AttachPropertyFromTargetToSource):
    _need_changes = frozenset(
        {
            str(IODDExtractor.__name__),
        }
    )

    def __init__(self):
        super().__init__(
            target_node_type=IODD.TextObject,
            target_property=IODD.value,
            delete_target_node=True,
            namespace=IODD,
        )


class IODDPruneDanglingNodes(PruneDanglingNodes):
    _need_changes = frozenset({str(IODDExtractor.__name__), str(IODDAttachPropertyFromTargetToSource.__name__)})

    def __init__(self):
        super().__init__(node_prune_types=[IODD.TextObject])



"""All warnings raised by the neat package are defined here. Note this module is called 'warnings' which
conflicts with the built-in Python warnings module. However, it is expected to always be used in an absolute
import, and should thus not cause a naming conflict."""

from cognite.neat._issues._base import NeatWarning, _get_subclasses

from . import user_modeling
from ._external import (
    CDFAuthWarning,
    CDFMaxIterationsWarning,
    FileItemNotSupportedWarning,
    FileMissingRequiredFieldWarning,
    FileReadWarning,
    FileTypeUnexpectedWarning,
)
from ._general import MissingCogniteClientWarning, NeatValueWarning, NotSupportedWarning, RegexViolationWarning
from ._models import (
    BreakingModelingPrincipleWarning,
    CDFNotSupportedWarning,
    NotSupportedHasDataFilterLimitWarning,
    NotSupportedViewContainerLimitWarning,
    PrincipleMatchingSpaceAndVersionWarning,
    PrincipleOneModelOneSpaceWarning,
    PrincipleSolutionBuildsOnEnterpriseWarning,
    UndefinedViewWarning,
    UserModelingWarning,
)
from ._properties import (
    PropertyDataTypeConversionWarning,
    PropertyDefinitionDuplicatedWarning,
    PropertyDirectRelationLimitWarning,
    PropertyMultipleValueWarning,
    PropertyNotFoundWarning,
    PropertyOverwritingWarning,
    PropertyTypeNotSupportedWarning,
    PropertyValueTypeUndefinedWarning,
)
from ._resources import (
    ResourceNeatWarning,
    ResourceNotFoundWarning,
    ResourceRegexViolationWarning,
    ResourceRetrievalWarning,
    ResourcesDuplicatedWarning,
    ResourceTypeNotSupportedWarning,
    ResourceUnknownWarning,
)

__all__ = [
    "BreakingModelingPrincipleWarning",
    "CDFAuthWarning",
    "CDFMaxIterationsWarning",
    "CDFNotSupportedWarning",
    "FileItemNotSupportedWarning",
    "FileMissingRequiredFieldWarning",
    "FileReadWarning",
    "FileTypeUnexpectedWarning",
    "MissingCogniteClientWarning",
    "NeatValueWarning",
    "NotSupportedHasDataFilterLimitWarning",
    "NotSupportedViewContainerLimitWarning",
    "NotSupportedWarning",
    "PrincipleMatchingSpaceAndVersionWarning",
    "PrincipleOneModelOneSpaceWarning",
    "PrincipleSolutionBuildsOnEnterpriseWarning",
    "PropertyDataTypeConversionWarning",
    "PropertyDefinitionDuplicatedWarning",
    "PropertyDirectRelationLimitWarning",
    "PropertyMultipleValueWarning",
    "PropertyNotFoundWarning",
    "PropertyOverwritingWarning",
    "PropertyTypeNotSupportedWarning",
    "PropertyValueTypeUndefinedWarning",
    "RegexViolationWarning",
    "ResourceNeatWarning",
    "ResourceNotFoundWarning",
    "ResourceRegexViolationWarning",
    "ResourceRetrievalWarning",
    "ResourceTypeNotSupportedWarning",
    "ResourceUnknownWarning",
    "ResourcesDuplicatedWarning",
    "UndefinedViewWarning",
    "UserModelingWarning",
    "user_modeling",
]

_NEAT_WARNINGS_BY_NAME = {warning.__name__: warning for warning in _get_subclasses(NeatWarning, include_base=True)}



import warnings
import zipfile
from collections.abc import Collection
from pathlib import Path
from typing import Any, Literal, overload

from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import DataModelIdentifier

from cognite.neat._alpha import AlphaFlags
from cognite.neat._constants import COGNITE_MODELS
from cognite.neat._graph import loaders
from cognite.neat._issues import IssueList, NeatIssue, catch_issues
from cognite.neat._rules import exporters
from cognite.neat._rules._constants import PATTERNS
from cognite.neat._rules._shared import VerifiedRules
from cognite.neat._rules.exporters._rules2dms import Component
from cognite.neat._rules.importers import DMSImporter
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.models.dms import DMSMetadata
from cognite.neat._utils.upload import UploadResultList

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class ToAPI:
    """API used to write the contents of a NeatSession to a specified destination. For instance writing information
    rules or DMS rules to a NEAT rules Excel spreadsheet, or writing a verified data model to CDF.

    """

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.cdf = CDFToAPI(state, verbose)
        self._python = ToPythonAPI(state, verbose)

    def ontology(self, io: Any) -> None:
        """Export the data model to ontology.

        Args:
            io: The file path to file-like object to write the session to.

        Example:
            Export the session to a file
            ```python
            ontology_file_name = "neat_session.ttl"
            neat.to.ontology(ontology_file_name)
            ```
        """
        warnings.filterwarnings("default")
        AlphaFlags.to_ontology.warn()

        filepath = Path(io)
        if filepath.suffix != ".ttl":
            warnings.warn("File extension is not .ttl, adding it to the file name", stacklevel=2)
            filepath = filepath.with_suffix(".ttl")

        exporter = exporters.OWLExporter()
        self._state.rule_store.export_to_file(exporter, Path(io))
        return None

    def excel(
        self,
        io: Any,
        include_reference: bool | DataModelIdentifier = True,
        include_properties: Literal["same-space", "all"] = "all",
        add_empty_rows: bool = False,
    ) -> IssueList | None:
        """Export the verified data model to Excel.

        Args:
            io: The file path or file-like object to write the Excel file to.
            include_reference: If True, the reference data model will be included. Defaults to True.
                Note that this only applies if you have created the data model using the
                create.enterprise_model(...), create.solution_model(), or create.data_product_model() methods.
                You can also provide a DataModelIdentifier directly, which will be read from CDF
            include_properties: The properties to include in the Excel file. Defaults to "all".
                - "same-space": Only properties that are in the same space as the data model will be included.
            add_empty_rows: If True, empty rows will be added between each component. Defaults to False.

        Example:
            Export information model to excel rules sheet
            ```python
            information_rules_file_name = "information_rules.xlsx"
            neat.to.excel(information_rules_file_name)
            ```

        Example:
            Read CogniteCore model, convert it to an enterprise model, and export it to an excel file
            ```python
            client = CogniteClient()
            neat = NeatSession(client)

            neat.read.cdf(("cdf_cdm", "CogniteCore", "v1"))
            neat.create.enterprise_model(
                data_model_id=("sp_doctrino_space", "ExtensionCore", "v1"),
                org_name="MyOrg",
            )
            dms_rules_file_name = "dms_rules.xlsx"
            neat.to.excel(dms_rules_file_name, include_reference=True)
            ```

        Example:
            Read the data model ("my_space", "ISA95Model", "v5") and export it to an excel file with the
            CogniteCore model in the reference sheets.
            ```python
            client = CogniteClient()
            neat = NeatSession(client)

            neat.read.cdf(("my_space", "ISA95Model", "v5"))
            dms_rules_file_name = "dms_rules.xlsx"
            neat.to.excel(dms_rules_file_name, include_reference=("cdf_cdm", "CogniteCore", "v1"))
        """
        reference_rules_with_prefix: tuple[VerifiedRules, str] | None = None
        include_properties = include_properties.strip().lower()

        if include_reference is not False:
            if include_reference is True and self._state.last_reference is not None:
                ref_rules: InformationRules | DMSRules | None = self._state.last_reference
            elif include_reference is True:
                ref_rules = None
            else:
                if not self._state.client:
                    raise NeatSessionError("No client provided!")
                ref_rules = None
                with catch_issues() as issues:
                    ref_read = DMSImporter.from_data_model_id(self._state.client, include_reference).to_rules()
                    if ref_read.rules is not None:
                        ref_rules = ref_read.rules.as_verified_rules()
                if ref_rules is None or issues.has_errors:
                    issues.action = f"Read {include_reference}"
                    return issues
            if ref_rules is not None:
                prefix = "Ref"
                if (
                    isinstance(ref_rules.metadata, DMSMetadata)
                    and ref_rules.metadata.as_data_model_id() in COGNITE_MODELS
                ):
                    prefix = "CDM"
                reference_rules_with_prefix = ref_rules, prefix

        if include_properties == "same-space":
            warnings.filterwarnings("default")
            AlphaFlags.same_space_properties_only_export.warn()

        exporter = exporters.ExcelExporter(
            styling="maximal",
            reference_rules_with_prefix=reference_rules_with_prefix,
            add_empty_rows=add_empty_rows,
            include_properties=include_properties,  # type: ignore
        )
        self._state.rule_store.export_to_file(exporter, Path(io))
        return None

    def session(self, io: Any) -> None:
        """Export the current session to a file.

        Args:
            io: The file path to file-like object to write the session to.

        Example:
            Export the session to a file
            ```python
            session_file_name = "neat_session.zip"
            neat.to.session(session_file_name)
            ```
        """
        filepath = Path(io)
        if filepath.suffix not in {".zip"}:
            warnings.warn("File extension is not .zip, adding it to the file name", stacklevel=2)
            filepath = filepath.with_suffix(".zip")

        filepath.parent.mkdir(exist_ok=True, parents=True)

        with zipfile.ZipFile(filepath, "w") as zip_ref:
            zip_ref.writestr(
                "neat-session/instances/instances.trig",
                self._state.instances.store.serialize(),
            )

    @overload
    def yaml(
        self,
        io: None,
        format: Literal["neat"] = "neat",
        skip_system_spaces: bool = True,
    ) -> str: ...

    @overload
    def yaml(
        self,
        io: Any,
        format: Literal["neat", "toolkit"] = "neat",
        skip_system_spaces: bool = True,
    ) -> None: ...

    def yaml(
        self, io: Any | None = None, format: Literal["neat", "toolkit"] = "neat", skip_system_spaces: bool = True
    ) -> str | None:
        """Export the verified data model to YAML.

        Args:
            io: The file path or file-like object to write the YAML file to. Defaults to None.
            format: The format of the YAML file. Defaults to "neat".
            skip_system_spaces: If True, system spaces will be skipped. Defaults to True.

        !!! note "YAML formats"
            - "neat": This is the format Neat uses to store the data model.
            - "toolkit": This is the format used by Cognite Toolkit, that matches the CDF API.

        Returns:
            str | None: If io is None, the YAML string will be returned. Otherwise, None will be returned.

        Example:
            Export to yaml file in the case of "neat" format
            ```python
            your_yaml_file_name = "neat_rules.yaml"
            neat.to.yaml(your_yaml_file_name, format="neat")
            ```

        Example:
            Export yaml files as a zip folder in the case of "toolkit" format
            ```python
            your_zip_folder_name = "toolkit_data_model_files.zip"
            neat.to.yaml(your_zip_folder_name, format="toolkit")
            ```

        Example:
            Export yaml files to a folder in the case of "toolkit" format
            ```python
            your_folder_name = "my_project/data_model_files"
            neat.to.yaml(your_folder_name, format="toolkit")
            ```
        """

        if format == "neat":
            exporter = exporters.YAMLExporter()
            if io is None:
                return self._state.rule_store.export(exporter)

            self._state.rule_store.export_to_file(exporter, Path(io))
        elif format == "toolkit":
            if io is None or not isinstance(io, str | Path):
                raise NeatSessionError(
                    "Please provide a zip file or directory path to write the YAML files to."
                    "This is required for the 'toolkit' format."
                )
            user_path = Path(io)
            if user_path.suffix == "" and not user_path.exists():
                user_path.mkdir(parents=True)
            self._state.rule_store.export_to_file(
                exporters.DMSExporter(remove_cdf_spaces=skip_system_spaces), user_path
            )
        else:
            raise NeatSessionError("Please provide a valid format. 'neat' or 'toolkit'")

        return None


@session_class_wrapper
class CDFToAPI:
    """Write a verified Data Model and Instances to CDF."""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def instances(
        self,
        space: str | None = None,
        space_property: str | None = None,
    ) -> UploadResultList:
        """Export the verified DMS instances to CDF.

        Args:
            space: Name of instance space to use. Default is to suffix the schema space with '_instances'.
                Note this space is required to be different from the space with the data model.
            space_property: This is an alternative to the 'space' argument. If provided, the space will set to the
                value of the property with the given name for each instance. If the property is not found, the
                'space' argument will be used. Defaults to None.

        Returns:
            UploadResultList: The result of the upload.

        Example:
            Export instances to CDF
            ```python
            neat.to.cdf.instances()
            ```

            Export instances to CDF using the `dataSetId` property as the space
            ```python
            neat.to.cdf.instances(space_property="dataSetId")
            ```

        """
        return self._instances(instance_space=space, space_from_property=space_property)

    def _instances(
        self,
        instance_space: str | None = None,
        space_from_property: str | None = None,
        use_source_space: bool = False,
    ) -> UploadResultList:
        if not self._state.client:
            raise NeatSessionError("No CDF client provided!")
        client = self._state.client
        dms_rules = self._state.rule_store.last_verified_dms_rules
        instance_space = instance_space or f"{dms_rules.metadata.space}_instances"

        if instance_space and instance_space == dms_rules.metadata.space:
            raise NeatSessionError("Space for instances must be different from the data model space.")
        elif not PATTERNS.space_compliance.match(str(instance_space)):
            raise NeatSessionError("Please provide a valid space name. {PATTERNS.space_compliance.pattern}")

        if not client.data_modeling.spaces.retrieve(instance_space):
            client.data_modeling.spaces.apply(dm.SpaceApply(space=instance_space))

        loader = loaders.DMSLoader(
            self._state.rule_store.last_verified_dms_rules,
            self._state.rule_store.last_verified_information_rules,
            self._state.instances.store,
            instance_space=instance_space,
            client=client,
            space_property=space_from_property,
            use_source_space=use_source_space,
            # In case urllib.parse.quote() was run on the extraction, we need to run
            # urllib.parse.unquote() on the load.
            unquote_external_ids=True,
            neat_prefix_by_predicate_uri=self._state.instances.neat_prefix_by_predicate_uri,
            neat_prefix_by_type_uri=self._state.instances.neat_prefix_by_type_uri,
        )

        result = loader.load_into_cdf(client)
        self._state.instances.outcome.append(result)
        print("You can inspect the details with the .inspect.outcome.instances(...) method.")
        return result

    def data_model(
        self,
        existing: Literal["fail", "skip", "update", "force", "recreate"] = "update",
        dry_run: bool = False,
        drop_data: bool = False,
        components: Component | Collection[Component] | None = None,
    ) -> UploadResultList:
        """Export the verified DMS data model to CDF.

        Args:
            existing: What to do if the component already exists. Defaults to "update".
                See the note below for more information about the options.
            dry_run: If True, no changes will be made to CDF. Defaults to False.
            drop_data: If existing is 'force' or 'recreate' and the operation will lead to data loss,
                the component will be skipped unless drop_data is True. Defaults to False.
                Note this only applies to spaces and containers if they contain data.
            components: The components to export. If None, all components will be exported. Defaults to None.

        !!! note "Data Model creation modes"
            - "fail": If any component already exists, the export will fail.
            - "skip": If any component already exists, it will be skipped.
            - "update": If any component already exists, it will be updated.
            - "force": If any component already exists, and the update fails, it will be deleted and recreated.
            - "recreate": All components will be deleted and recreated. The exception is spaces, which will be updated.

        """

        exporter = exporters.DMSExporter(existing=existing, export_components=components, drop_data=drop_data)

        if not self._state.client:
            raise NeatSessionError("No client provided!")

        result = self._state.rule_store.export_to_cdf(exporter, self._state.client, dry_run)
        print("You can inspect the details with the .inspect.outcome.data_model(...) method.")
        return result


@session_class_wrapper
class ToPythonAPI:
    """API used to write the contents of a NeatSession to Python objects"""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def instances(
        self,
        instance_space: str | None = None,
        space_from_property: str | None = None,
        use_source_space: bool = False,
    ) -> tuple[list[dm.InstanceApply], IssueList]:
        """Export the verified DMS instances to Python objects.

        Args:
            instance_space: The name of the instance space to use. Defaults to None.
            space_from_property: This is an alternative to the 'instance_space' argument. If provided,
                the space will be set to the value of the property with the given name for each instance.
                If the property is not found, the 'instance_space' argument will be used. Defaults to None.
            use_source_space: If True, the instance space will be set to the source space of the instance.
                This is only relevant if the instances were extracted from CDF data models. Defaults to False.

        Returns:
            list[dm.InstanceApply]: The instances as Python objects.

        Example:
            Export instances to Python objects
            ```python
            instances = neat.to._python.instances()
            ```

            Export instances to Python objects using the `dataSetId` property as the space
            ```python
            instances = neat.to._python.instances(space_from_property="dataSetId")
            ```
        """
        dms_rules = self._state.rule_store.last_verified_dms_rules
        instance_space = instance_space or f"{dms_rules.metadata.space}_instances"

        if instance_space and instance_space == dms_rules.metadata.space:
            raise NeatSessionError("Space for instances must be different from the data model space.")
        elif not PATTERNS.space_compliance.match(str(instance_space)):
            raise NeatSessionError(f"Please provide a valid space name. {PATTERNS.space_compliance.pattern}")

        loader = loaders.DMSLoader(
            self._state.rule_store.last_verified_dms_rules,
            self._state.rule_store.last_verified_information_rules,
            self._state.instances.store,
            instance_space=instance_space,
            space_property=space_from_property,
            use_source_space=use_source_space,
            unquote_external_ids=True,
            neat_prefix_by_predicate_uri=self._state.instances.neat_prefix_by_predicate_uri,
            neat_prefix_by_type_uri=self._state.instances.neat_prefix_by_type_uri,
        )
        issue_list = IssueList()
        instances: list[dm.InstanceApply] = []
        for item in loader.load(stop_on_exception=False):
            if isinstance(item, dm.InstanceApply):
                instances.append(item)
            elif isinstance(item, NeatIssue):
                issue_list.append(item)
        return instances, issue_list



import urllib.parse
from collections.abc import Iterable, Set
from typing import Any, cast

from cognite.client.data_classes import Row, RowList
from cognite.client.utils.useful_types import SequenceNotStr
from rdflib import RDF, Namespace, URIRef

from cognite.neat._client import NeatClient
from cognite.neat._constants import DEFAULT_RAW_URI
from cognite.neat._shared import Triple

from ._base import BaseExtractor
from ._dict import DEFAULT_EMPTY_VALUES, DictExtractor


class RAWExtractor(BaseExtractor):
    def __init__(
        self,
        client: NeatClient,
        db_name: str,
        table_name: str,
        table_type: str | None = None,
        foreign_keys: str | SequenceNotStr[str] | None = None,
        namespace: Namespace | None = None,
        empty_values: Set[str] = DEFAULT_EMPTY_VALUES,
        str_to_ideal_type: bool = False,
        unpack_json: bool = False,
    ) -> None:
        self.client = client
        self.db_name = db_name
        self.table_name = table_name
        self.table_type = table_type
        self.foreign_keys = {foreign_keys} if isinstance(foreign_keys, str) else set(foreign_keys or [])
        self.namespace = namespace or Namespace(DEFAULT_RAW_URI)
        self.empty_values = empty_values
        self.str_to_ideal_type = str_to_ideal_type
        self.unpack_json = unpack_json

    @property
    def _rdf_type(self) -> URIRef:
        return self.namespace[urllib.parse.quote(self.table_type or self.table_name)]

    def extract(self) -> Iterable[Triple]:
        for row in self.client.raw.rows(self.db_name, self.table_name, partitions=10, chunk_size=None):
            if isinstance(row, Row):
                yield from self._row2triples(row)
            elif isinstance(row, RowList):
                # Bug in SDK returning row list with chunk_size= None
                for item in row:
                    yield from self._row2triples(item)

    def _row2triples(self, row: Row) -> Iterable[Triple]:
        # The row is always set. It is just the PySDK that have it as str | None
        key, data = cast(tuple[str, dict[str, Any]], (row.key, row.columns))
        identifier = self.namespace[urllib.parse.quote(key)]
        yield identifier, RDF.type, self._rdf_type

        yield from DictExtractor(
            identifier,
            data,
            self.namespace,
            self.foreign_keys,
            self.empty_values,
            self.str_to_ideal_type,
            self.unpack_json,
        ).extract()



from ._rules import InformationClass, InformationMetadata, InformationProperty, InformationRules
from ._rules_input import (
    InformationInputClass,
    InformationInputMetadata,
    InformationInputProperty,
    InformationInputRules,
)
from ._validation import InformationValidation

__all__ = [
    "InformationClass",
    "InformationInputClass",
    "InformationInputMetadata",
    "InformationInputProperty",
    "InformationInputRules",
    "InformationMetadata",
    "InformationProperty",
    "InformationRules",
    "InformationValidation",
]



from ._base import BaseTransformerStandardised
from ._classic_cdf import (
    AddAssetDepth,
    AssetEventConnector,
    AssetFileConnector,
    AssetRelationshipConnector,
    AssetSequenceConnector,
    AssetTimeSeriesConnector,
    LookupRelationshipSourceTarget,
    RelationshipAsEdgeTransformer,
)
from ._prune_graph import (
    AttachPropertyFromTargetToSource,
    PruneDanglingNodes,
    PruneDeadEndEdges,
    PruneInstancesOfUnknownType,
    PruneTypes,
)
from ._rdfpath import MakeConnectionOnExactMatch
from ._value_type import ConnectionToLiteral, ConvertLiteral, LiteralToEntity, SetType, SplitMultiValueProperty

__all__ = [
    "AddAssetDepth",
    "AssetEventConnector",
    "AssetFileConnector",
    "AssetRelationshipConnector",
    "AssetSequenceConnector",
    "AssetTimeSeriesConnector",
    "AttachPropertyFromTargetToSource",
    "ConnectionToLiteral",
    "ConvertLiteral",
    "LiteralToEntity",
    "LookupRelationshipSourceTarget",
    "MakeConnectionOnExactMatch",
    "PruneDanglingNodes",
    "PruneDeadEndEdges",
    "PruneInstancesOfUnknownType",
    "PruneTypes",
    "RelationshipAsEdgeTransformer",
    "SetType",
    "SplitMultiValueProperty",
]

Transformers = (
    AddAssetDepth
    | AssetTimeSeriesConnector
    | AssetSequenceConnector
    | AssetFileConnector
    | AssetEventConnector
    | AssetRelationshipConnector
    | SplitMultiValueProperty
    | RelationshipAsEdgeTransformer
    | MakeConnectionOnExactMatch
    | AttachPropertyFromTargetToSource
    | PruneDanglingNodes
    | PruneTypes
    | PruneDeadEndEdges
    | PruneInstancesOfUnknownType
    | ConvertLiteral
    | LiteralToEntity
    | ConnectionToLiteral
    | BaseTransformerStandardised
    | LookupRelationshipSourceTarget
    | SetType
)



from typing import cast

import pandas as pd
from rdflib import URIRef

from cognite.neat._utils.rdf_ import remove_namespace_from_uri
from cognite.neat._utils.text import humanize_collection

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class ExploreAPI:
    """
    Explore the instances in the session.
    """

    def __init__(self, state: SessionState):
        self._state = state

    def types(self) -> pd.DataFrame:
        """List all the types of instances in the session."""
        return pd.DataFrame(self._state.instances.store.queries.types_with_instance_and_property_count())

    def properties(self) -> pd.DataFrame:
        """List all the properties of a type of instances in the session."""
        return pd.DataFrame(self._state.instances.store.queries.properties_with_count())

    def instance_with_properties(self, type: str) -> dict[str, set[str]]:
        """List all the instances of a type with their properties."""
        available_types = self._state.instances.store.queries.list_types(remove_namespace=False)
        uri_by_type = {remove_namespace_from_uri(t[0]): t[0] for t in available_types}
        if type not in uri_by_type:
            raise NeatSessionError(
                f"Type {type} not found. Available types are: {humanize_collection(uri_by_type.keys())}"
            )
        type_uri = cast(URIRef, uri_by_type[type])
        return self._state.instances.store.queries.instances_with_properties(type_uri, remove_namespace=True)



import warnings
from collections.abc import Hashable
from typing import TYPE_CHECKING, Any, ClassVar, Literal

import pandas as pd
from cognite.client import data_modeling as dm
from pydantic import Field, field_serializer, field_validator, model_validator
from pydantic_core.core_schema import SerializationInfo, ValidationInfo

from cognite.neat._client.data_classes.schema import DMSSchema
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.warnings._general import NeatValueWarning
from cognite.neat._rules.models._base_rules import (
    BaseMetadata,
    BaseRules,
    ContainerProperty,
    DataModelAspect,
    RoleTypes,
    SheetList,
    SheetRow,
    ViewProperty,
    ViewRef,
)
from cognite.neat._rules.models._types import (
    ClassEntityType,
    ContainerEntityType,
    DmsPropertyType,
    StrListType,
    URIRefType,
    ViewEntityType,
)
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import (
    ContainerEntityList,
    DMSEntity,
    DMSNodeEntity,
    DMSUnknownEntity,
    EdgeEntity,
    Entity,
    HasDataFilter,
    NodeTypeFilter,
    RawFilter,
    ReverseConnectionEntity,
    ViewEntity,
    ViewEntityList,
)

if TYPE_CHECKING:
    from cognite.neat._rules.models import InformationRules

_DEFAULT_VERSION = "1"


class DMSMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.dms
    aspect: ClassVar[DataModelAspect] = DataModelAspect.physical
    logical: URIRefType | None = None

    def as_space(self) -> dm.SpaceApply:
        return dm.SpaceApply(
            space=self.space,
        )

    def as_data_model(self) -> dm.DataModelApply:
        suffix = f"Creator: {', '.join(self.creator)}"
        if self.description:
            description = f"{self.description} Creator: {', '.join(self.creator)}"
        else:
            description = suffix

        return dm.DataModelApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name or None,
            version=self.version or "missing",
            description=description,
            views=[],
        )

    def get_prefix(self) -> str:
        return self.space


def _metadata(context: Any) -> DMSMetadata | None:
    if isinstance(context, dict) and isinstance(context.get("metadata"), DMSMetadata):
        return context["metadata"]
    return None


class DMSProperty(SheetRow):
    view: ViewEntityType = Field(alias="View", description="The property identifier.")
    view_property: DmsPropertyType = Field(alias="View Property", description="The ViewId this property belongs to")
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the property")
    description: str | None = Field(alias="Description", default=None, description="Short description of the property")
    connection: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None = Field(
        None,
        alias="Connection",
        description="nly applies to connection between views. "
        "It specify how the connection should be implemented in CDF.",
    )
    value_type: DataType | ViewEntity | DMSUnknownEntity = Field(
        alias="Value Type",
        description="Value type that the property can hold. It takes either subset of CDF primitive types or a View id",
    )
    nullable: bool | None = Field(
        default=None,
        alias="Nullable",
        description="Used to indicate whether the property is required or not. Only applies to primitive type.",
    )
    immutable: bool | None = Field(
        default=None,
        alias="Immutable",
        description="sed to indicate whether the property is can only be set once. Only applies to primitive type.",
    )
    is_list: bool | None = Field(
        default=None,
        alias="Is List",
        description="Used to indicate whether the property holds single or multiple values (list). "
        "Only applies to primitive types.",
    )
    default: bool | str | int | float | dict | None = Field(
        None, alias="Default", description="Specifies default value for the property."
    )
    container: ContainerEntityType | None = Field(
        None,
        alias="Container",
        description="Specifies container where the property is stored. Only applies to primitive type.",
    )
    container_property: DmsPropertyType | None = Field(
        None,
        alias="Container Property",
        description="Specifies property in the container where the property is stored. Only applies to primitive type.",
    )
    index: StrListType | None = Field(
        None,
        alias="Index",
        description="The names of the indexes (comma separated) that should be created for the property.",
    )
    constraint: StrListType | None = Field(
        None,
        alias="Constraint",
        description="The names of the uniquness (comma separated) that should be created for the property.",
    )
    logical: URIRefType | None = Field(
        None,
        alias="Logical",
        description="Used to make connection between physical and logical data model aspect",
    )

    def _identifier(self) -> tuple[Hashable, ...]:
        return self.view, self.view_property

    @field_validator("nullable")
    def direct_relation_must_be_nullable(cls, value: Any, info: ValidationInfo) -> None:
        if info.data.get("connection") == "direct" and value is False:
            raise ValueError("Direct relation must be nullable")
        return value

    @field_validator("value_type", mode="after")
    def connections_value_type(
        cls, value: EdgeEntity | ViewEntity | DMSUnknownEntity, info: ValidationInfo
    ) -> DataType | EdgeEntity | ViewEntity | DMSUnknownEntity:
        if (connection := info.data.get("connection")) is None:
            return value
        if connection == "direct" and not isinstance(value, ViewEntity | DMSUnknownEntity):
            raise ValueError(f"Direct relation must have a value type that points to a view, got {value}")
        elif isinstance(connection, EdgeEntity) and not isinstance(value, ViewEntity):
            raise ValueError(f"Edge connection must have a value type that points to a view, got {value}")
        elif isinstance(connection, ReverseConnectionEntity) and not isinstance(value, ViewEntity):
            raise ValueError(f"Reverse connection must have a value type that points to a view, got {value}")
        return value

    @field_validator("default", mode="after")
    def set_proper_type_on_default(cls, value: Any, info: ValidationInfo) -> Any:
        if not value:
            return value
        value_type = info.data.get("value_type")
        if not isinstance(value_type, DataType):
            warnings.filterwarnings("default")
            warnings.warn(
                NeatValueWarning(f"Default value {value} set to connection {value_type} will be ignored"),
                stacklevel=2,
            )
            return None
        else:
            try:
                return value_type.convert_value(value)
            except ValueError:
                warnings.filterwarnings("default")
                warnings.warn(
                    NeatValueWarning(f"Could not convert {value} to {value_type}"),
                    stacklevel=2,
                )
                return None

    @field_validator("container", "container_property", mode="after")
    def container_set_correctly(cls, value: Any, info: ValidationInfo) -> Any:
        if (connection := info.data.get("connection")) is None:
            return value
        if connection == "direct" and value is None:
            raise ValueError(
                "You must provide a container and container property for where to store direct connections"
            )
        elif isinstance(connection, EdgeEntity) and value is not None:
            raise ValueError(
                "Edge connections are not stored in a container, please remove the container and container property"
            )
        elif isinstance(connection, ReverseConnectionEntity) and value is not None:
            raise ValueError(
                "Reverse connection are not stored in a container, please remove the container and container property"
            )
        return value

    @field_serializer("value_type", when_used="always")
    def as_dms_type(self, value_type: DataType | EdgeEntity | ViewEntity, info: SerializationInfo) -> str:
        if isinstance(value_type, DataType):
            return value_type._suffix_extra_args(value_type.dms._type)
        elif isinstance(value_type, EdgeEntity | ViewEntity) and (metadata := _metadata(info.context)):
            return value_type.dump(space=metadata.space, version=metadata.version)
        return str(value_type)

    @field_serializer("view", "container", when_used="unless-none")
    def remove_default_space(self, value: str, info: SerializationInfo) -> str:
        if (metadata := _metadata(info.context)) and isinstance(value, Entity):
            if info.field_name == "container" and info.context.get("as_reference") is True:
                # When dumping as reference, the container should keep the default space for easy copying
                # over to user sheets.
                return value.dump()
            return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("connection", when_used="unless-none")
    def remove_defaults(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, Entity) and (metadata := _metadata(info.context)):
            default_type = f"{self.view.external_id}.{self.view_property}"
            if isinstance(value, EdgeEntity) and value.edge_type and value.edge_type.space != metadata.space:
                default_type = f"{metadata.space}{default_type}"
            return value.dump(space=metadata.space, version=metadata.version, type=default_type)
        return str(value)

    def as_container_reference(self) -> ContainerProperty:
        if self.container is None or self.container_property is None:
            raise NeatValueError("Accessing container reference without container and container property set")
        return ContainerProperty(container=self.container, property_=self.container_property)

    def as_view_reference(self) -> ViewProperty:
        return ViewProperty(view=self.view, property_=self.view_property)


class DMSContainer(SheetRow):
    container: ContainerEntityType = Field(
        alias="Container", description="Container id, strongly advised to PascalCase usage."
    )
    name: str | None = Field(
        alias="Name", default=None, description="Human readable name of the container being defined."
    )
    description: str | None = Field(
        alias="Description", default=None, description="Short description of the node being defined."
    )
    constraint: ContainerEntityList | None = Field(
        None, alias="Constraint", description="List of required (comma separated) constraints for the container"
    )
    used_for: Literal["node", "edge", "all"] | None = Field(
        "all", alias="Used For", description=" Whether the container is used for nodes, edges or all."
    )

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.container,)

    def as_container(self) -> dm.ContainerApply:
        container_id = self.container.as_id()
        constraints: dict[str, dm.Constraint] = {}
        for constraint in self.constraint or []:
            requires = dm.RequiresConstraint(constraint.as_id())
            constraints[f"{constraint.space}_{constraint.external_id}"] = requires

        return dm.ContainerApply(
            space=container_id.space,
            external_id=container_id.external_id,
            name=self.name or None,
            description=self.description,
            constraints=constraints or None,
            properties={},
            used_for=self.used_for,
        )

    @field_serializer("container", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if metadata := _metadata(info.context):
            if isinstance(value, DMSEntity):
                return value.dump(space=metadata.space, version=metadata.version)
            elif isinstance(value, Entity):
                return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("constraint", when_used="unless-none")
    def remove_default_spaces(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, list) and (metadata := _metadata(info.context)):
            return ",".join(
                constraint.dump(space=metadata.space, version=metadata.version)
                if isinstance(constraint, DMSEntity)
                else str(constraint)
                for constraint in value
            )
        return ",".join(str(value) for value in value)


class DMSView(SheetRow):
    view: ViewEntityType = Field(alias="View", description="View id, strongly advised to PascalCase usage.")
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the view being defined.")
    description: str | None = Field(
        alias="Description", default=None, description="Short description of the view being defined "
    )
    implements: ViewEntityList | None = Field(
        None,
        alias="Implements",
        description="List of parent view ids (comma separated) which the view being defined implements.",
    )
    filter_: HasDataFilter | NodeTypeFilter | RawFilter | None = Field(
        None, alias="Filter", description="Explicitly define the filter for the view."
    )
    in_model: bool = Field(
        True,
        alias="In Model",
        description="Indicates whether the view being defined is a part of the data model.",
    )
    logical: URIRefType | None = Field(
        None,
        alias="Logical",
        description="Used to make connection between physical and logical data model aspect",
    )

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.view,)

    @field_serializer("view", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if (metadata := _metadata(info.context)) and isinstance(value, Entity):
            return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("implements", when_used="unless-none")
    def remove_default_spaces(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, list) and (metadata := _metadata(info.context)):
            return ",".join(
                parent.dump(space=metadata.space, version=metadata.version)
                if isinstance(parent, DMSEntity)
                else str(parent)
                for parent in value
            )
        return ",".join(str(value) for value in value) if isinstance(value, list) else value

    def as_view(self) -> dm.ViewApply:
        view_id = self.view.as_id()
        implements = [parent.as_id() for parent in self.implements or []] or None

        return dm.ViewApply(
            space=view_id.space,
            external_id=view_id.external_id,
            version=view_id.version or _DEFAULT_VERSION,
            name=self.name or None,
            description=self.description,
            filter=None if self.filter_ is None else self.filter_.as_dms_filter(),
            implements=implements,
            properties={},
        )

    def as_view_reference(self) -> ViewRef:
        return ViewRef(view=self.view)


class DMSNode(SheetRow):
    node: DMSNodeEntity = Field(alias="Node", description="The type definition of the node.")
    usage: Literal["type", "collection"] = Field(
        alias="Usage", description="What the usage of the node is in the data model."
    )
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the node being defined.")
    description: str | None = Field(
        alias="Description", default=None, description="Short description of the node being defined."
    )

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.node,)

    def as_node(self) -> dm.NodeApply:
        if self.usage == "type":
            return dm.NodeApply(space=self.node.space, external_id=self.node.external_id)
        elif self.usage == "collection":
            raise NotImplementedError("Collection nodes are not supported yet")
        else:
            raise ValueError(f"Unknown usage {self.usage}")

    @field_serializer("node", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, DMSEntity) and (metadata := _metadata(info.context)):
            return value.dump(space=metadata.space, version=metadata.version)
        return str(value)


class DMSEnum(SheetRow):
    collection: ClassEntityType = Field(alias="Collection", description="The collection this enum belongs to.")
    value: str = Field(alias="Value", description="The value of the enum.")
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the enum.")
    description: str | None = Field(alias="Description", default=None, description="Short description of the enum.")

    def _identifier(self) -> tuple[Hashable, ...]:
        return self.collection, self.value

    @field_serializer("collection", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, DMSEntity) and (metadata := _metadata(info.context)):
            return value.dump(space=metadata.space, version=metadata.version)
        return str(value)


class DMSRules(BaseRules):
    metadata: DMSMetadata = Field(alias="Metadata", description="Contains information about the data model.")
    properties: SheetList[DMSProperty] = Field(
        alias="Properties", description="Contains the properties of the data model."
    )
    views: SheetList[DMSView] = Field(alias="Views", description="Contains the views of the data model.")
    containers: SheetList[DMSContainer] | None = Field(
        None,
        alias="Containers",
        description="Contains the definition containers that are the physical storage of the data model.",
    )
    enum: SheetList[DMSEnum] | None = Field(None, alias="Enum", description="Contains the definition of enum values.")
    nodes: SheetList[DMSNode] | None = Field(
        None, alias="Nodes", description="Contains the definition of the node types."
    )

    @model_validator(mode="after")
    def set_neat_id(self) -> "DMSRules":
        namespace = self.metadata.namespace

        for view in self.views:
            if not view.neatId:
                view.neatId = namespace[view.view.suffix]

        for property_ in self.properties:
            if not property_.neatId:
                property_.neatId = namespace[f"{property_.view.suffix}/{property_.view_property}"]

        return self

    def update_neat_id(self) -> None:
        """Update neat ids"""

        namespace = self.metadata.namespace

        for view in self.views:
            view.neatId = namespace[view.view.suffix]

        for property_ in self.properties:
            property_.neatId = namespace[f"{property_.view.suffix}/{property_.view_property}"]

    def sync_with_info_rules(self, info_rules: "InformationRules") -> None:
        # Sync at the metadata level
        if info_rules.metadata.physical == self.metadata.identifier:
            self.metadata.logical = info_rules.metadata.identifier
        else:
            # if models are not linked to start with, we skip
            return None

        info_properties_by_neat_id = {prop.neatId: prop for prop in info_rules.properties}
        dms_properties_by_neat_id = {prop.neatId: prop for prop in self.properties}
        for neat_id, prop in info_properties_by_neat_id.items():
            if prop.physical in dms_properties_by_neat_id:
                dms_properties_by_neat_id[prop.physical].logical = neat_id

        info_classes_by_neat_id = {cls.neatId: cls for cls in info_rules.classes}
        dms_views_by_neat_id = {view.neatId: view for view in self.views}
        for neat_id, class_ in info_classes_by_neat_id.items():
            if class_.physical in dms_views_by_neat_id:
                dms_views_by_neat_id[class_.physical].logical = neat_id

    def as_schema(self, instance_space: str | None = None, remove_cdf_spaces: bool = False) -> DMSSchema:
        from ._exporter import _DMSExporter

        return _DMSExporter(self, instance_space, remove_cdf_spaces=remove_cdf_spaces).to_schema()

    @classmethod
    def display_type_name(cls) -> str:
        return "VerifiedDMSModel"

    def _repr_html_(self) -> str:
        summary = {
            "aspect": self.metadata.aspect,
            "intended for": "DMS Architect",
            "name": self.metadata.name,
            "space": self.metadata.space,
            "external_id": self.metadata.external_id,
            "version": self.metadata.version,
            "views": len(self.views),
            "containers": len(self.containers) if self.containers else 0,
            "properties": len(self.properties),
        }

        return pd.DataFrame([summary]).T.rename(columns={0: ""})._repr_html_()  # type: ignore



import itertools
import json
import urllib.parse
import warnings
from collections import defaultdict
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, cast, get_args, overload

import yaml
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.capabilities import Capability, DataModelInstancesAcl
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType
from cognite.client.data_classes.data_modeling.ids import InstanceId
from cognite.client.data_classes.data_modeling.views import SingleEdgeConnection
from cognite.client.exceptions import CogniteAPIError
from pydantic import BaseModel, ValidationInfo, create_model, field_validator
from rdflib import RDF, URIRef

from cognite.neat._client import NeatClient
from cognite.neat._client._api_client import SchemaAPI
from cognite.neat._constants import DMS_DIRECT_RELATION_LIST_LIMIT, is_readonly_property
from cognite.neat._issues import IssueList, NeatError, NeatIssue, catch_issues
from cognite.neat._issues.errors import (
    AuthorizationError,
    ResourceCreationError,
    ResourceDuplicatedError,
    ResourceNotFoundError,
)
from cognite.neat._issues.warnings import (
    NeatValueWarning,
    PropertyDirectRelationLimitWarning,
    PropertyMultipleValueWarning,
    PropertyTypeNotSupportedWarning,
    ResourceNeatWarning,
)
from cognite.neat._rules.analysis import RulesAnalysis
from cognite.neat._rules.analysis._base import ViewQuery, ViewQueryDict
from cognite.neat._rules.models import DMSRules
from cognite.neat._rules.models.data_types import _DATA_TYPE_BY_DMS_TYPE, Json, String
from cognite.neat._rules.models.information._rules import InformationRules
from cognite.neat._shared import InstanceType
from cognite.neat._store import NeatGraphStore
from cognite.neat._utils.auxiliary import create_sha256_hash
from cognite.neat._utils.collection_ import iterate_progress_bar_if_above_config_threshold
from cognite.neat._utils.rdf_ import namespace_as_space, remove_namespace_from_uri, split_uri
from cognite.neat._utils.text import NamingStandardization
from cognite.neat._utils.upload import UploadResult

from ._base import _END_OF_CLASS, _START_OF_CLASS, CDFLoader


@dataclass
class _ViewIterator:
    """This is a helper class to iterate over the views

    Args:
        view_id: The view to iterate over
        instance_count: The number of instances in the view
        query: The query to get the instances from the store.
        view: The view object from the client.
    """

    view_id: dm.ViewId
    instance_count: int
    query: ViewQuery
    view: dm.View | None = None


@dataclass
class _Projection:
    """This is a helper class to project triples to a node and/or edge(s)"""

    view_id: dm.ViewId
    used_for: Literal["node", "edge", "all"]
    pydantic_cls: type[BaseModel]
    edge_by_type: dict[str, tuple[str, dm.EdgeConnection]]
    edge_by_prop_id: dict[str, tuple[str, dm.EdgeConnection]]


class DMSLoader(CDFLoader[dm.InstanceApply]):
    """Loads Instances to Cognite Data Fusion Data Model Service from NeatGraph.

    Args:
        dms_rules (DMSRules): The DMS rules used by the data model.
        info_rules (InformationRules): The information rules used by the data model, used to
            look+up the instances in the store.
        graph_store (NeatGraphStore): The graph store to load the data from.
        instance_space (str): The instance space to load the data into.
        create_issues (Sequence[NeatIssue] | None): A list of issues that occurred during reading. Defaults to None.
        client (NeatClient | None): This is used to lookup containers such that the loader
            creates instances in accordance with required constraints. Defaults to None.
        unquote_external_ids (bool): If True, the loader will unquote external ids before creating the instances.
        neat_prefix_by_predicate_uri (dict[URIRef, str] | None): A dictionary that maps a predicate URIRef to a
            prefix that Neat added to the object upon extraction. This is used to remove the prefix from the
            object before creating the instance.
    """

    def __init__(
        self,
        dms_rules: DMSRules,
        info_rules: InformationRules,
        graph_store: NeatGraphStore,
        instance_space: str,
        space_property: str | None = None,
        use_source_space: bool = False,
        client: NeatClient | None = None,
        create_issues: Sequence[NeatIssue] | None = None,
        unquote_external_ids: bool = False,
        neat_prefix_by_predicate_uri: dict[URIRef, str] | None = None,
        neat_prefix_by_type_uri: dict[URIRef, str] | None = None,
    ):
        super().__init__(graph_store)
        self.dms_rules = dms_rules
        self.info_rules = info_rules
        self.neat_prefix_by_predicate_uri = neat_prefix_by_predicate_uri or {}
        self.neat_prefix_by_type_uri = neat_prefix_by_type_uri or {}
        self._instance_space = instance_space
        self._space_property = space_property
        self._use_source_space = use_source_space
        self._space_by_instance_uri: dict[URIRef, str] = defaultdict(lambda: instance_space)
        self._external_id_by_uri: dict[URIRef, str] = {}
        self._issues = IssueList(create_issues or [])
        self._client = client
        self._unquote_external_ids = unquote_external_ids

    def write_to_file(self, filepath: Path) -> None:
        if filepath.suffix not in [".json", ".yaml", ".yml"]:
            raise ValueError(f"File format {filepath.suffix} is not supported")
        dumped: dict[str, list] = {"nodes": [], "edges": [], "issues": []}
        for item in self.load(stop_on_exception=False):
            key = {
                dm.NodeApply: "nodes",
                dm.EdgeApply: "edges",
                NeatIssue: "issues",
            }.get(type(item))
            if key is None:
                # This should never happen, and is a bug in neat
                raise ValueError(f"Item {item} is not supported. This is a bug in neat please report it.")
            dumped[key].append(item.dump())
        with filepath.open("w", encoding=self._encoding, newline=self._new_line) as f:
            if filepath.suffix == ".json":
                json.dump(dumped, f, indent=2)
            else:
                yaml.safe_dump(dumped, f, sort_keys=False)

    def _load(
        self, stop_on_exception: bool = False
    ) -> Iterable[dm.InstanceApply | NeatIssue | type[_END_OF_CLASS] | _START_OF_CLASS]:
        if self._issues.has_errors and stop_on_exception:
            raise self._issues.as_exception()
        elif self._issues.has_errors:
            yield from self._issues
            return
        view_iterations, issues = self._create_view_iterations()
        yield from issues
        if self._space_property:
            yield from self._lookup_space_by_uri(view_iterations, stop_on_exception)

            if self._client:
                space_creation = self._create_instance_space_if_not_exists()
                yield from space_creation.warnings
                if space_creation.has_errors and stop_on_exception:
                    raise space_creation.as_exception()
                yield from space_creation.errors

        if self.neat_prefix_by_type_uri:
            self._lookup_identifier_by_uri()

        for it in view_iterations:
            view = it.view
            if view is None:
                yield ResourceNotFoundError(it.view_id, "view", more=f"Skipping {it.instance_count} instances...")
                continue
            projection, issues = self._create_projection(view)
            yield from issues
            query = it.query
            reader = self.graph_store.read(
                query.rdf_type,
                property_renaming_config=query.property_renaming_config,
                remove_uri_namespace=False,
            )
            instance_iterable = iterate_progress_bar_if_above_config_threshold(
                reader, it.instance_count, f"Loading {it.view_id!r}"
            )
            yield _START_OF_CLASS(view.external_id)
            for identifier, properties in instance_iterable:
                yield from self._create_instances(identifier, properties, projection, stop_on_exception)
            if reader is instance_iterable:
                print(f"Loaded {it.instance_count} instances for {it.view_id!r}")

            yield _END_OF_CLASS

    def _create_view_iterations(self) -> tuple[list[_ViewIterator], IssueList]:
        view_query_by_id = RulesAnalysis(self.info_rules, self.dms_rules).view_query_by_id
        iterations_by_view_id = self._select_views_with_instances(view_query_by_id)
        if self._client:
            issues = IssueList()
            views = self._client.data_modeling.views.retrieve(
                list(iterations_by_view_id.keys()), include_inherited_properties=True
            )
            if missing := set(iterations_by_view_id) - {view.as_id() for view in views}:
                for missing_view in missing:
                    issues.append(ResourceNotFoundError(missing_view, "view", more="The view is not found in CDF."))
                return [], issues
        else:
            views = dm.ViewList([])
            with catch_issues() as issues:
                read_model = self.dms_rules.as_schema().as_read_model()
                views.extend(read_model.views)
            if issues.has_errors:
                return [], issues
        views_by_id = {view.as_id(): view for view in views}

        def sort_by_instance_type(id_: dm.ViewId) -> int:
            if id_ not in views_by_id:
                return 0
            return {"node": 1, "all": 1, "edge": 3}.get(views_by_id[id_].used_for, 0)

        ordered_view_ids = SchemaAPI.get_view_order_by_direct_relation_constraints(views)
        # Sort is stable in Python, so we will keep the order of the views:
        ordered_view_ids.sort(key=sort_by_instance_type)
        view_iterations: list[_ViewIterator] = []
        for view_id in ordered_view_ids:
            if view_id not in iterations_by_view_id:
                continue
            view_iteration = iterations_by_view_id[view_id]
            view_iteration.view = views_by_id.get(view_id)
            view_iterations.append(view_iteration)
        return view_iterations, issues

    def _select_views_with_instances(self, view_query_by_id: ViewQueryDict) -> dict[dm.ViewId, _ViewIterator]:
        """Selects the views with data."""
        view_iterations: dict[dm.ViewId, _ViewIterator] = {}
        for view_id, query in view_query_by_id.items():
            count = self.graph_store.queries.count_of_type(query.rdf_type)
            if count > 0:
                view_iterations[view_id] = _ViewIterator(view_id, count, query)
        return view_iterations

    def _lookup_space_by_uri(self, view_iterations: list[_ViewIterator], stop_on_exception: bool = False) -> IssueList:
        issues = IssueList()
        if self._space_property is None:
            return issues
        total = sum(it.instance_count for it in view_iterations)
        properties_by_uriref = self.graph_store.queries.properties()
        space_property_uri = next((k for k, v in properties_by_uriref.items() if v == self._space_property), None)
        if space_property_uri is None:
            error: ResourceNotFoundError[str, str] = ResourceNotFoundError(
                self._space_property,
                "property",
                more=f"Could not find the {self._space_property} in the graph.",
            )
            if stop_on_exception:
                raise error
            issues.append(error)
            return issues

        instance_iterable = self.graph_store.queries.list_instances_ids_by_space(space_property_uri)
        instance_iterable = iterate_progress_bar_if_above_config_threshold(
            instance_iterable, total, f"Looking up spaces for {total} instances..."
        )
        neat_prefix = self.neat_prefix_by_predicate_uri.get(space_property_uri)
        warned_spaces: set[str] = set()
        for instance, space in instance_iterable:
            if neat_prefix:
                space = space.removeprefix(neat_prefix)

            clean_space = NamingStandardization.standardize_space_str(space)
            if clean_space != space and space not in warned_spaces:
                issues.append(
                    NeatValueWarning(
                        f"Invalid space in property {self._space_property}: {space}. Fixed to {clean_space}"
                    )
                )
                warned_spaces.add(space)

            self._space_by_instance_uri[instance] = clean_space
        return issues

    def _lookup_identifier_by_uri(self) -> None:
        if not self.neat_prefix_by_type_uri:
            return

        count = sum(count for _, count in self.graph_store.queries.summarize_instances())
        instance_iterable = self.graph_store.queries.list_instances_ids()
        instance_iterable = iterate_progress_bar_if_above_config_threshold(
            instance_iterable, count, f"Looking up identifiers for {count} instances..."
        )
        count_by_identifier: dict[str, list[URIRef]] = defaultdict(list)
        for instance_uri, type in instance_iterable:
            if type not in self.neat_prefix_by_type_uri:
                continue
            prefix = self.neat_prefix_by_type_uri[type]
            identifier = remove_namespace_from_uri(instance_uri)
            if self._unquote_external_ids:
                identifier = urllib.parse.unquote(identifier)
            count_by_identifier[identifier.removeprefix(prefix)].append(instance_uri)

        for identifier, uris in count_by_identifier.items():
            if len(uris) == 1:
                self._external_id_by_uri[uris[0]] = identifier

    def _create_instance_space_if_not_exists(self) -> IssueList:
        issues = IssueList()
        if not self._client:
            return issues

        instance_spaces = set(self._space_by_instance_uri.values()) - {self._instance_space}
        existing_spaces = {space.space for space in self._client.data_modeling.spaces.retrieve(list(instance_spaces))}
        if missing_spaces := (instance_spaces - existing_spaces):
            try:
                self._client.data_modeling.spaces.apply([dm.SpaceApply(space=space) for space in missing_spaces])
            except CogniteAPIError as e:
                issues.append(AuthorizationError(f"Creating {len(missing_spaces)} instance spaces.", str(e)))
        return issues

    def _create_projection(self, view: dm.View) -> tuple[_Projection, IssueList]:
        issues = IssueList()
        field_definitions: dict[str, tuple[type, Any]] = {}
        edge_by_type: dict[str, tuple[str, dm.EdgeConnection]] = {}
        edge_by_prop_id: dict[str, tuple[str, dm.EdgeConnection]] = {}
        validators: dict[str, classmethod] = {}
        direct_relation_by_property: dict[str, dm.DirectRelation] = {}
        unit_properties: list[str] = []
        json_fields: list[str] = []
        text_fields: list[str] = []
        for prop_id, prop in view.properties.items():
            if isinstance(prop, dm.EdgeConnection):
                if prop.edge_source:
                    # Edges with properties are created separately
                    continue

                edge_by_type[prop.type.external_id] = prop_id, prop
                edge_by_prop_id[prop_id] = prop_id, prop

            if isinstance(prop, dm.MappedProperty):
                if is_readonly_property(prop.container, prop.container_property_identifier):
                    continue

                if isinstance(prop.type, dm.DirectRelation):
                    if prop.container == dm.ContainerId("cdf_cdm", "CogniteTimeSeries") and prop_id == "unit":
                        unit_properties.append(prop_id)
                    else:
                        direct_relation_by_property[prop_id] = prop.type
                    python_type: Any = dict
                else:
                    data_type = _DATA_TYPE_BY_DMS_TYPE.get(prop.type._type)
                    if not data_type:
                        issues.append(
                            PropertyTypeNotSupportedWarning(
                                view.as_id(),
                                "view",
                                prop_id,
                                prop.type._type,
                            )
                        )
                        continue

                    if data_type == Json:
                        json_fields.append(prop_id)
                    elif data_type == String:
                        text_fields.append(prop_id)
                    python_type = data_type.python
                if isinstance(prop.type, ListablePropertyType) and prop.type.is_list:
                    python_type = list[python_type]
                default_value: Any = prop.default_value
                if prop.nullable:
                    python_type = python_type | None
                else:
                    default_value = ...

                field_definitions[prop_id] = (python_type, default_value)

        def parse_list(cls, value: Any, info: ValidationInfo) -> list[str]:
            if isinstance(value, list) and list.__name__ not in _get_field_value_types(cls, info):
                if len(value) > 1:
                    warnings.warn(
                        # the identifier is unknown, it will be cest in the create_instances method
                        PropertyMultipleValueWarning("", "property", str(info.field_name), value=str(value[0])),
                        stacklevel=2,
                    )
                elif not value:
                    return None  # type: ignore[return-value]
                return value[0]

            return value

        def parse_json_string(cls, value: Any, info: ValidationInfo) -> dict | list:
            if isinstance(value, dict):
                return value
            elif isinstance(value, list):
                try:
                    return [json.loads(v) if isinstance(v, str) else v for v in value]
                except json.JSONDecodeError as error:
                    raise ValueError(f"Not valid JSON string for {info.field_name}: {value}, error {error}") from error
            elif isinstance(value, str):
                try:
                    return json.loads(value)
                except json.JSONDecodeError as error:
                    raise ValueError(f"Not valid JSON string for {info.field_name}: {value}, error {error}") from error
            else:
                raise ValueError(f"Expect valid JSON string or dict for {info.field_name}: {value}")

        if json_fields:
            validators["parse_json_string"] = field_validator(*json_fields, mode="before")(parse_json_string)  # type: ignore[assignment, arg-type]

        validators["parse_list"] = field_validator("*", mode="before")(parse_list)  # type: ignore[assignment, arg-type]

        if direct_relation_by_property:

            def parse_direct_relation(cls, value: list, info: ValidationInfo) -> dict | list[dict]:
                # We validate above that we only get one value for single direct relations.
                if list.__name__ in _get_field_value_types(cls, info):
                    ids = (self._create_instance_id(v, "node", stop_on_exception=True) for v in value)
                    result = [id_.dump(camel_case=True, include_instance_type=False) for id_ in ids]
                    # Todo: Account for max_list_limit
                    if len(result) <= DMS_DIRECT_RELATION_LIST_LIMIT:
                        return result
                    warnings.warn(
                        PropertyDirectRelationLimitWarning(
                            identifier="unknown",
                            resource_type="view property",
                            property_name=cast(str, cls.model_fields[info.field_name].alias or info.field_name),
                            limit=DMS_DIRECT_RELATION_LIST_LIMIT,
                        ),
                        stacklevel=2,
                    )
                    # To get deterministic results, we sort by space and externalId
                    result.sort(key=lambda x: (x["space"], x["externalId"]))
                    return result[:DMS_DIRECT_RELATION_LIST_LIMIT]
                elif value:
                    return self._create_instance_id(value[0], "node", stop_on_exception=True).dump(
                        camel_case=True, include_instance_type=False
                    )
                return {}

            validators["parse_direct_relation"] = field_validator(*direct_relation_by_property.keys(), mode="before")(  # type: ignore[assignment]
                parse_direct_relation  # type: ignore[arg-type]
            )

        if unit_properties:

            def parse_direct_relation_to_unit(cls, value: Any, info: ValidationInfo) -> dict | list[dict]:
                if value:
                    external_id = remove_namespace_from_uri(value[0])
                    if self._unquote_external_ids:
                        external_id = urllib.parse.unquote(external_id)
                    return {"space": "cdf_cdm_units", "externalId": external_id}
                return {}

            validators["parse_direct_relation_to_unit"] = field_validator(*unit_properties, mode="before")(  # type: ignore[assignment]
                parse_direct_relation_to_unit  # type: ignore[arg-type]
            )

        if text_fields:

            def parse_text(cls, value: Any, info: ValidationInfo) -> str | list[str]:
                if isinstance(value, list):
                    return [remove_namespace_from_uri(v) if isinstance(v, URIRef) else str(v) for v in value]
                return remove_namespace_from_uri(value) if isinstance(value, URIRef) else str(value)

            validators["parse_text"] = field_validator(*text_fields, mode="before")(parse_text)  # type: ignore[assignment, arg-type]

        pydantic_cls = create_model(view.external_id, __validators__=validators, **field_definitions)  # type: ignore[arg-type, call-overload]
        return _Projection(view.as_id(), view.used_for, pydantic_cls, edge_by_type, edge_by_prop_id), issues

    def _create_instances(
        self,
        instance_uri: URIRef,
        properties: dict[str | InstanceType, list[Any]],
        projection: _Projection,
        stop_on_exception: Literal[True, False] = False,
    ) -> Iterable[dm.InstanceApply | NeatIssue]:
        instance_id = self._create_instance_id(instance_uri, "node", stop_on_exception)
        if not isinstance(instance_id, InstanceId):
            yield instance_id
            return
        space, external_id = instance_id.space, instance_id.external_id
        start_node, end_node = self._pop_start_end_node(properties)
        is_edge = start_node and end_node
        instance_type = "edge" if is_edge else "node"
        if (projection.used_for == "node" and is_edge) or (projection.used_for == "edge" and not is_edge):
            creation_error = ResourceCreationError(
                external_id,
                instance_type,
                f"View used for {projection.used_for} instance {external_id!s} but is {instance_type}",
            )
            if stop_on_exception:
                raise creation_error from None
            yield creation_error
            return

        if RDF.type not in properties:
            error = ResourceCreationError(external_id, instance_type, "No rdf:type found")
            if stop_on_exception:
                raise error from None
            yield error
            return
        _ = properties.pop(RDF.type)[0]

        sources = []
        with catch_issues() as property_issues:
            sources = [
                dm.NodeOrEdgeData(
                    projection.view_id,
                    projection.pydantic_cls.model_validate(properties).model_dump(
                        exclude_unset=True,
                        exclude_none=True,
                    ),
                )
            ]
        for issue in property_issues:
            if isinstance(issue, ResourceNeatWarning):
                issue.identifier = external_id

        if property_issues.has_errors and stop_on_exception:
            raise property_issues.as_exception()
        yield from property_issues
        if not sources:
            return

        if start_node and end_node:
            start = self._create_instance_id(start_node, "edge", stop_on_exception)
            end = self._create_instance_id(end_node, "edge", stop_on_exception)
            if isinstance(start, NeatError):
                yield start
            if isinstance(end, NeatError):
                yield end
            if isinstance(start, InstanceId) and isinstance(end, InstanceId):
                yield dm.EdgeApply(
                    space=space,
                    external_id=external_id,
                    type=(projection.view_id.space, projection.view_id.external_id),
                    start_node=start.as_tuple(),
                    end_node=end.as_tuple(),
                    sources=sources,
                )
        else:
            yield dm.NodeApply(
                space=space,
                external_id=external_id,
                type=(projection.view_id.space, projection.view_id.external_id),
                sources=sources,
            )
        yield from self._create_edges_without_properties(space, external_id, properties, projection, stop_on_exception)

    def _create_edges_without_properties(
        self,
        space: str,
        identifier: str,
        properties: dict[str | InstanceType, list[str] | list[URIRef]],
        projection: _Projection,
        stop_on_exception: Literal[True, False],
    ) -> Iterable[dm.EdgeApply | NeatIssue]:
        for predicate, values in properties.items():
            if predicate in projection.edge_by_type:
                prop_id, edge = projection.edge_by_type[predicate]
            elif predicate in projection.edge_by_prop_id:
                prop_id, edge = projection.edge_by_prop_id[predicate]
            else:
                continue
            if isinstance(edge, SingleEdgeConnection) and len(values) > 1:
                error = ResourceDuplicatedError(
                    resource_type="edge",
                    identifier=identifier,
                    location=f"Multiple values for single edge {edge}. Expected only one.",
                )
                yield error
                continue
            for target in values:
                target_id = self._create_instance_id(target, "edge", stop_on_exception)  # type: ignore[call-overload]
                if not isinstance(target_id, InstanceId):
                    yield target_id
                    continue
                if isinstance(target, URIRef):
                    target = remove_namespace_from_uri(target)
                external_id = f"{identifier}.{prop_id}.{target}"

                start_node, end_node = (
                    (space, identifier),
                    target_id.as_tuple(),
                )
                if edge.direction == "inwards":
                    start_node, end_node = end_node, start_node
                yield dm.EdgeApply(
                    space=space,
                    external_id=(external_id if len(external_id) < 256 else create_sha256_hash(external_id)),
                    type=edge.type,
                    start_node=start_node,
                    end_node=end_node,
                )

    @staticmethod
    def _pop_start_end_node(
        properties: dict[str | InstanceType, list[str] | list[URIRef]],
    ) -> tuple[URIRef, URIRef] | tuple[None, None]:
        start_node = properties.pop("startNode", [None])[0]
        if not start_node:
            start_node = properties.pop("start_node", [None])[0]
        end_node = properties.pop("endNode", [None])[0]
        if not end_node:
            end_node = properties.pop("end_node", [None])[0]
        if start_node and end_node:
            return start_node, end_node  # type: ignore[return-value]
        return None, None

    @overload
    def _create_instance_id(
        self, uri: URIRef, instance_type: str, stop_on_exception: Literal[False] = False
    ) -> InstanceId | NeatError: ...

    @overload
    def _create_instance_id(
        self, uri: URIRef, instance_type: str, stop_on_exception: Literal[True] = True
    ) -> InstanceId: ...

    def _create_instance_id(
        self, uri: URIRef, instance_type: str, stop_on_exception: bool = False
    ) -> InstanceId | NeatError:
        space: str | None = None
        external_id: str | None = None
        error: NeatError | None = None
        if self._use_source_space:
            namespace, external_id = split_uri(uri)
            space = namespace_as_space(namespace)
            if space is None:
                error = ResourceCreationError(uri, instance_type, f"Could not find space for {uri!s}.")
        else:
            space = self._space_by_instance_uri[uri]
            if uri in self._external_id_by_uri:
                external_id = self._external_id_by_uri[uri]
            else:
                external_id = remove_namespace_from_uri(uri)

        if external_id and self._unquote_external_ids:
            external_id = urllib.parse.unquote(external_id)
        if space and external_id:
            return InstanceId(space, external_id)
        if error is None:
            raise ValueError(f"Bug in neat. Failed to create instance ID and determine error for {uri!r}")
        if stop_on_exception:
            raise error
        return error

    def _get_required_capabilities(self) -> list[Capability]:
        return [
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Write,
                    DataModelInstancesAcl.Action.Write_Properties,
                    DataModelInstancesAcl.Action.Read,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID([self._instance_space]),
            )
        ]

    def _upload_to_cdf(
        self,
        client: CogniteClient,
        items: list[dm.InstanceApply],
        dry_run: bool,
        read_issues: IssueList,
        class_name: str | None = None,
    ) -> Iterable[UploadResult]:
        name = class_name or "Instances"
        nodes = [item for item in items if isinstance(item, dm.NodeApply)]
        edges = [item for item in items if isinstance(item, dm.EdgeApply)]
        try:
            upserted = client.data_modeling.instances.apply(
                nodes,
                edges,
                auto_create_end_nodes=True,
                auto_create_start_nodes=True,
                skip_on_version_conflict=True,
            )
        except CogniteAPIError as e:
            if len(items) == 1:
                yield UploadResult(
                    name=name,
                    issues=read_issues,
                    failed_items=items,
                    error_messages=[str(e)],
                    failed_upserted={item.as_id() for item in items},  # type: ignore[attr-defined]
                )
            else:
                half = len(items) // 2
                yield from self._upload_to_cdf(client, items[:half], dry_run, read_issues, class_name)
                yield from self._upload_to_cdf(client, items[half:], dry_run, read_issues, class_name)
        else:
            result = UploadResult(name=name, issues=read_issues)  # type: ignore[var-annotated]
            for instance in itertools.chain(upserted.nodes, upserted.edges):  # type: ignore[attr-defined]
                if instance.was_modified and instance.created_time == instance.last_updated_time:
                    result.created.add(instance.as_id())
                elif instance.was_modified:
                    result.changed.add(instance.as_id())
                else:
                    result.unchanged.add(instance.as_id())
            yield result


def _get_field_value_types(cls, info):
    return [type_.__name__ for type_ in get_args(cls.model_fields[info.field_name].annotation)]
